// FILE: Game.java

/* GEMpire (Generic Empire for the Masses), a java package to support internet oriented  
 * empire-like strategy games and associated utilities. 
 * Copyright (C) 1998 Christopher L. Marshall 
 * 
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 *  
 * My current email is christopherlmarshall@yahoo.com.  My home address is: 
 * 10006 Leamoore Lane, 
 * Vienna VA, 22181 
 *  
 * It is my hope that this modest effort might assist others in producing Java 
 * programs to implement a varied host of empire-like games. 
 */ package GEMpire;

import java.io.*;

/**  This class stores player specific information not associated with the map, like whether she
  *  has moved this turn or not.
  */
class PlayerInfo{
   // instance variables
   boolean     finishedCurrentTurn= false;
   int         lastLoginNumber    = -1;
   boolean     loggedIn           = false;
   boolean     annihilated        = false;
   String      password           = "pass";
   PrintWriter logfile;

   // constructors
   PlayerInfo(String filename) throws IOException{
      logfile= new PrintWriter(new FileWriter(filename));
   }
   PlayerInfo(BufferedReader in, String filename) throws IOException{
      finishedCurrentTurn= new Boolean(in.readLine().trim()).booleanValue(); 
      annihilated= new Boolean(in.readLine().trim()).booleanValue(); 
      password= in.readLine().trim();
      logfile= IOUtil.openPrintWriterToAppend(filename);
   }

   // read/write
   void write(PrintWriter out){
      out.println(finishedCurrentTurn);
      out.println(annihilated);
      out.println(password);
   }
   void openNewLogfile(String filename) throws IOException{
      logfile= new PrintWriter(new FileWriter(filename));
   }

   // actions
   public void implementNewTurn(){ 
      finishedCurrentTurn= false; 
   } 
}

/* This class manages info about the players not associated with the map, like whether a player has
 * moved or not this turn, if a given player number is valid in this game, who controls the 'board'...
 */

public class Game{
   // instance variables
   private PlayerInfo  pinfo[]; 
   private int         currentTurn=1;
   private int         playerInControl= -1;
   private int         nextLoginNumber= 0;
   private boolean     hasAnyoneLoggedInYet= false;
   private PrintWriter playerLogFiles[];

   // constructors
   public Game(boolean players[]){ 
      pinfo= new PlayerInfo[players.length]; 
      for (int i=0; i<pinfo.length; i++){ 
         if (players[i]){ 
            try { 
               pinfo[i]= new PlayerInfo(getFilenameForTurn(currentTurn,i)); 
            } 
            catch (IOException e){ 
               Globals.exit("Game.init(): could not open log file for player "+i+"!\n\r"+e); 
            } 
         } 
         else pinfo[i]= null; 
      } 
   } 
 
   public Game(BufferedReader in) throws IOException{ 
      if (IOUtil.skipPastLine("GameFileBegin",in)){ 
         boolean done; 
          
         // Get number of possible players and allocate pinfo accordingly. 
         pinfo= new PlayerInfo[Integer.parseInt(in.readLine().trim())]; 
         for (int i=0; i<pinfo.length; i++) pinfo[i]= null; 
 
         currentTurn= Integer.parseInt(in.readLine().trim()); 
         hasAnyoneLoggedInYet= new Boolean(in.readLine().trim()).booleanValue(); 
 
         // Read individual player info. 
         int numberOfPlayers= Integer.parseInt(in.readLine().trim()); 
         for (int i=0; i<numberOfPlayers; i++){ 
            int p= Integer.parseInt(in.readLine().trim()); 
            try { 
               pinfo[p]= new PlayerInfo( 
                  in, 
                  getFilenameForTurn(currentTurn,i) 
               ); 
            } 
            catch (IOException e){ 
               Globals.exit("Game.init(): could not open log file for player "+i+"!\n\r"+e); 
            } 
         } 
      } 
      else throw new IOException("Game.init(): file improperly formatted, did not contain Game info"); 
   } 

   // get/set

   public void setPassword(int playerNumber, String password){
      pinfo[playerNumber].password= password;
   }
   public void killPlayer(int i){
      pinfo[i].annihilated= true;
   }
   public PrintWriter getLogfile(int i){ 
      return pinfo[i].logfile; 
   } 
   public String getFilenameForTurn(int turn, int player){ 
      return "player"+player+"Turn"+turn+"Log.txt"; 
   } 
   private int numberOfValidPlayers(){ 
      int tmp=0; 
      for (int i=0; i<pinfo.length; i++){ 
         if (pinfo[i]!=null) if (!pinfo[i].annihilated){ 
            tmp++; 
         } 
      } 
      return tmp; 
   } 
   public boolean validPlayer(int i){ 
      if (i<0||i>=pinfo.length) return false; 
      if (pinfo[i]!=null){
         if (!pinfo[i].annihilated) return true;
      }
      return false;
   } 
   public int getCurrentTurn(){return currentTurn;} 
   public int getControllingPlayer(){return playerInControl;} 
   public boolean isBeginning(){return !hasAnyoneLoggedInYet;} 
   public boolean isLoggedIn(int i){ 
      if (pinfo[i]==null) return false; 
      return pinfo[i].loggedIn; 
   } 
   public boolean controlsBoard(int i){ 
      return (playerInControl==i); 
   } 
   public boolean everyoneHasFinishedTurn(){ 
      // we skip player 0 because it is not a real player
      // but allows remote control of the server.
      for (int i=1; i<pinfo.length; i++){ 
         if (pinfo[i]!=null){ 
            if (!pinfo[i].annihilated&&!pinfo[i].finishedCurrentTurn) return false; 
         } 
      } 
      return true; 
   } 

   // actions

   /** Close log files for last turn and open files for current turn. */
   private void refreshLogfiles(){
      for (int i=0; i<pinfo.length; i++){
         PlayerInfo pi= pinfo[i];
         if (pi!=null){
            pi.logfile.close();
            try {
               pi.openNewLogfile(getFilenameForTurn(currentTurn,i));
            }
            catch (IOException e){
               Globals.exit(
                  "Game.refreshLogfiles(): could not open log file for player "+
                  i+"!\n\r"+
                  e
               );
            }
         }
      }
   }

   public void implementNewTurn(){
      currentTurn++;
      refreshLogfiles();
      for (int i=0; i<pinfo.length; i++){
         if (pinfo[i]!=null)
            pinfo[i].implementNewTurn();
      }
      passControlToNextInLine();
   }

   // login/board control related

   public boolean loginPlayer(int i, String password){
// this is obselete now.
//      if (!validPlayer(i)) 
//         Globals.exit("Game.loginPlayer(): "+i+" is not a valid player.  This shouldn't happen!");

      // check password.
      if (!password.equals(pinfo[i].password)) return false;

      if (playerInControl==-1){
         if (!pinfo[i].finishedCurrentTurn){
            if (validPlayer(i)){
               playerInControl= i;
            }
         }
      }
      pinfo[i].lastLoginNumber= nextLoginNumber++;
      pinfo[i].loggedIn= true;
      hasAnyoneLoggedInYet= true;

      return true;
   }

   public void logoffPlayer(int i){
/* this can happen now that players can be annihilated
      if (!validPlayer(i)) 
         Globals.exit("Game.logoutPlayer(): "+i+" is not a valid player.  This shouldn't happen!");
*/

      pinfo[i].loggedIn= false;
      if (playerInControl==i){
         passControlToNextInLine();
      }
   }

   /** This routine figures out who controls the board.  Calling it twice in a row will result in the same player 
     * controlling the board after the second call as controlled the board after the first call.  In other words, 
     * it never kicks anyone out, it just determines who should be in control, which may be the same player 
     * as before the call. 
     */ 
   private void passControlToNextInLine(){ 
      int tmp= -1; 
      int j= -1; 
      // Find the player who logged in first among those players presently logged it, assign to tmp. 
      // notice that we skip player 0.  That's because player 0 is not a regular
      // player but a device to implement remote control of the server
      // for the game referee.
      for (int i=1; i<pinfo.length; i++){ 
         if (pinfo[i]!=null){ 
            PlayerInfo pi= pinfo[i]; 
            if (!pi.finishedCurrentTurn&&pi.loggedIn&&!pi.annihilated){ 
               if (pi.lastLoginNumber<tmp||tmp==-1){ 
                  tmp= pi.lastLoginNumber; 
                  j=i; 
               } 
            } 
         } 
      } 
 
      if (tmp==-1) playerInControl= -1; else playerInControl= j; 
   } 
 
   public boolean endTurn(int i){ 
      boolean wasInControl= controlsBoard(i); 
System.out.println("Game.endTurn("+i+")"); 
      pinfo[i].finishedCurrentTurn= true; 
      if (playerInControl==i){ 
         passControlToNextInLine(); 
      } 
      return wasInControl; 
   } 
 
   // read/write

   public void write(PrintWriter out){
      out.println("GameFileBegin");
      out.println(pinfo.length);
      out.println(""+currentTurn);
      out.println(""+hasAnyoneLoggedInYet);
      out.println(numberOfValidPlayers());
      for (int i=0; i<pinfo.length; i++){
         if (pinfo[i]!=null){
            out.println(i);
            pinfo[i].write(out);
         }
      }
      out.println("GameFileEnd");
   }

   public static void main(String args[]) throws Exception{
      boolean players[]={
         true,
         true,
         true,
         true,
         true,
         false,
         false,
         true,
         false,
         false,
         false,
         false,
         false,
         false,
         false,
         false
      };

      PrintWriter pw= new PrintWriter(new FileWriter("test.out"));
      Game g= new Game(players);
      g.write(pw);
      pw.close();
      BufferedReader br= new BufferedReader(new FileReader("test.out"));
      Game g2= new Game(br); br.close();
      pw= new PrintWriter(new FileWriter("test2.out"));
      g2.write(pw);
      pw.close();
   }
}

