//FILE: GenMap.java 
 
/* GEMpire (Generic Empire for the Masses), a java package to support internet oriented  
 * empire-like strategy games and associated utilities. 
 * Copyright (C) 1998 Christopher L. Marshall 
 * 
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details. 
 *  
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. 
 *  
 * My current email is christopherlmarshall@yahoo.com.  My home address is: 
 * 10006 Leamoore Lane, 
 * Vienna VA, 22181 
 *  
 * It is my hope that this modest effort might assist others in producing Java 
 * programs to implement a varied host of empire-like games. 
 */ 

/**
  *  This class is for random map generation.
  *  It uses its own MapCell class and grid[][], instead of using the main Map class 
  *  directly.  The reason for this is that map generation requires special hooks
  *  and concepts, like region markers to keep adjacent land masses from merging
  *  as they grow together, that would clutter the main Map class, if added there.
  *
  */
 
package GEMpire;

import java.io.*;
import java.awt.Frame;
import java.util.*;

public class GenMap{
   static int defaultTerrain=TerrainTypes.SEA;

   // These two variables are something of a kludge.  Piece types are defined in a text file that is loaded on
   // startup.  Most GEMpire code is agnostic on assigning meaning to piece types beyond what is implied by
   // the model of possible piece properties represented by the PieceDatabase class.  That is, the GEMpire code isn't
   // supposed to think of cities per se, but it knows that some pieces can produce pieces of other types because
   // of the existence of the 'boolean canProduce(int typea, int typeb)' method of the PieceDatabase class.
   // So here we are in a class whose purpose is to generate random maps and we are hard coding values from 
   // a configuration file.  If you think about it, however, map generation can't be agnostic about piece properties
   // and what they mean.  If you change the piece info text file too much (the file that defines all piece types and
   // what their properties are), you need a different map generation algorithm.
   static final int neutralPlayerNumber=0;
   static final int cityPieceType=0;

   static class MapCell{
      int terrain=TerrainTypes.SEA;
      int region=0;
      MapCell(int terrain){
         this.terrain= terrain;
      }
      void set(int t, int r){
         terrain= t;
         region= r;
      }
   }

   class TerrainMap{

   // This inner class partitions all cells of the same terrain in a map
   // into non-touching regions.  This has many uses, but the most pressing
   // one is to figure out where the external sea is so that when scattering
   // cities on the map, each continent is guaranteed to have at least one
   // city that can make troop transports that can reach other land masses.
   // Here is  a PDL description of the algorithm used in the constructor:
   // 
   // get a list of all cells with the given terrain and mark them as in region '0',
   //    which means they don't belong to a region
   //    this is the 'unmarked' list referred to below
   // 
   // while (unmarked cells left){
   //    get unmarked cell
   //    mark cell
   //    remove from unmarked list
   // 
   //    start new region list
   //    put cell in region list
   //    put cell in working list
   // 
   //    while (working list not empty){
   //       remove first cell
   //       for each neighbor {
   //          if (inbounds and terrain matches)
   //             if (region==0){ // i.e. unmarked
   //                mark it
   //                remove from unmarked list
   //                add to region list
   //                add to working list
   //             }
   //             else if (region==working region)
   //                do nothing
   //                this cell was dealt with in a previous pass
   //             }
   //             else {
   //               this shouldn't happen
   //               report flaw in algorithm
   //             }
   //          }
   //       }
   //    }
   // }

      Hashtable unmarked= new Hashtable();
      Vector regions= new Vector();
      int terrain;

      // used mainly for debugging the TerrainMap class
      void write(PrintStream out){
         out.println(regions.size()+" regions");
         for (int i=0; i<regions.size(); i++){
            Vector region= (Vector)regions.elementAt(i);
            out.println("region "+i);
            for (int j=0; j<region.size(); j++){
               int rc[]= (int[])region.elementAt(j);
               out.println("("+rc[0]+","+rc[1]+")");
            }
         }
      }

      TerrainMap(int terrain){
         this.terrain= terrain;

         // gather list of all cells for the given terrain
         // and mark them as in region 0 (which is considered
         // 'unmarked')
         for (int r=0; r<numRows; r++){
            for (int c=0; c<numCols; c++){
               if (getTerrain(r,c)==terrain){
                  int rc[]= new int[2];
                  rc[0]= r; rc[1]= c;
                  unmarked.put(r+","+c,rc);
                  set(rc,terrain,0);
               }
            }
         }

         Enumeration keys= unmarked.keys();
         int N= unmarked.size();
         int region= 0;
         Vector curRegion= new Vector();
         int numNeighbors= MapHex.numNeighbors();
         while (keys.hasMoreElements()){
            // get first cell and remove from hashtable
            Object key= keys.nextElement();
            int rc[]= (int [])unmarked.get(key);
            unmarked.remove(key);

            // start a new region
            region++;
            Vector curRegionList= new Vector();
            regions.addElement(curRegionList);

            // mark this cell
            set(rc,terrain,region);

            // add it to the current region
            curRegionList.addElement(rc);

            // start working list and add cell to it.
            // the working list is the list of cells in the current
            // region that haven't had all their neighbors examined yet
            Vector workingList= new Vector();
            workingList.addElement(rc);
             
            while (workingList.size()>0){
               // remove first element from working list
               rc= (int[])workingList.elementAt(0);
               workingList.removeElementAt(0);

               for (int i=0; i<numNeighbors; i++){
                  int rc2[]= MapHex.getNeighborCoordinates(i,rc[0],rc[1]);
                  if (inBounds(rc2)){
                     if (getTerrain(rc2)==terrain){
                        int reg2= getRegion(rc2);
                        if (reg2==0){
                           // mark neighbor
                           set(rc2,terrain,region);

                           // remove it from the unmarked list
                           unmarked.remove(rc2[0]+","+rc2[1]);

                           // add it to the current region list
                           curRegionList.addElement(rc2);

                           // and add it to the working list
                           // so we'll check its neighbors later
                           workingList.addElement(rc2);
                        }
                        else if (reg2==region){
                           // do nothing.  This cell has already been added to the current
                           // region on a previous pass.
                        }
                        else {
                           System.err.println("This shouldn't happen!  Two regions are touching!");
                           System.exit(1);
                        }
                     }
                  }
               }
            }

            // get enumeration of remaining unmarked cells
            keys= unmarked.keys();
         }
      }
   }

   // instance variables for GenMap
   int numRows,numCols,numRegions;
   MapCell grid[][];
   Vector landCells;
   Vector seaCells;
   Map m= null;
   int externalSea;
   int numNeighbors= MapHex.numNeighbors();

   // constructors

   // this one reads an existing map from a file
   // only the terrain is read into the GenMap grid[][].
   public GenMap(String filename) throws 
      FileNotFoundException,
      IOException,
      BadRowAndColumnValuesException
   {
      FileReader fr= new FileReader(filename);
      Map m= new Map(null,fr);
      numRows= m.getNumRows();
      numCols= m.getNumCols();

      grid= new MapCell[numRows][numCols];
      for (int r=0; r<numRows; r++){
         for (int c=0; c<numCols; c++){
            grid[r][c]= new MapCell(m.getTerrain(r,c));
         }
      }
   }

   // My first generation algorithm to get things going.
   GenMap(int numRows, int numCols, int numSeeds, int numGrowSteps, int thresh){
      this.numRows= numRows;
      this.numCols= numCols;
      this.numRegions= numSeeds;

      grid= new MapCell[numRows][numCols];
      for (int r=0; r<numRows; r++){
         for (int c=0; c<numCols; c++){
            grid[r][c]= new MapCell(defaultTerrain);
         }
      }

      landCells= new Vector();

      // generate initial seeds
      int r,c;
      for (int i=0; i<numSeeds;){
         r= SystemUtil.rint(numRows);
         c= SystemUtil.rint(numCols);
         
         if (getTerrain(r,c)==TerrainTypes.SEA){
            if (countLandNeighbors(r,c)==0){
               int rc[]= new int[2];  rc[0]=r; rc[1]=c;
               landCells.addElement(rc);
               grid[r][c].set(TerrainTypes.LAND,i+1);
               i++;
            }
         }
      }

      // grow the landmasses
      for (int i=0; i<numGrowSteps; i++){
         growLandMasses(thresh);
      }
   }

   // my second algorithm to generate maps with more interesting controls
   GenMap(int numRows, int numCols, int numSeeds, double landRatio, int thresh, double quitDelta){
      this.numRows= numRows;
      this.numCols= numCols;
      this.numRegions= numSeeds;

      grid= new MapCell[numRows][numCols];
      for (int r=0; r<numRows; r++){
         for (int c=0; c<numCols; c++){
            grid[r][c]= new MapCell(defaultTerrain);
         }
      }

      landCells= new Vector();

      // generate initial seeds
      int r,c;
      for (int i=0; i<numSeeds;){
         r= SystemUtil.rint(numRows);
         c= SystemUtil.rint(numCols);
         
         if (getTerrain(r,c)==TerrainTypes.SEA){
            if (countLandNeighbors(r,c)==0){
               int rc[]= new int[2];  rc[0]=r; rc[1]=c;
               landCells.addElement(rc);
               grid[r][c].set(TerrainTypes.LAND,i+1);
               i++;
            }
         }
      }

      // grow the landmasses
      int lastN,curN,i=1;
      double lratio,qratio;
      curN= landCells.size();
      do {
         growLandMasses(thresh);

         lastN= curN;
         curN= landCells.size();
         qratio= (curN-lastN)/(1.0*lastN);
         lratio=curN/(numRows*numCols*1.0);
         System.err.println("iteration "+(i++)+" land ratio= "+lratio);
      } while (qratio>quitDelta && lratio<landRatio);
   }

   // tirival map information

   final void set(int rc[], int terrain, int region){grid[rc[0]][rc[1]].set(terrain,region);}
   final void set(int r, int c, int terrain, int region){grid[r][c].set(terrain,region);};
   final int getTerrain(int r, int c){return grid[r][c].terrain;}
   final int getTerrain(int rc[]){return grid[rc[0]][rc[1]].terrain;}
   final int getRegion(int r, int c){return grid[r][c].region;}
   final int getRegion(int rc[]){return grid[rc[0]][rc[1]].region;}
   final boolean inBounds(int rc[]){
      if ((0<=rc[0])&&(rc[0]<numRows)&&(rc[1]>=0)&&(rc[1]<numCols)){
         return true;
      }
      return false;
   }

   // non-trivial map information

   boolean connectsRegion(int r, int c, int region){
      int N= MapHex.numNeighbors();
      for (int i=0; i<N; i++){
         int rc[]= MapHex.getNeighborCoordinates(i,r,c);
         if (inBounds(rc)){
            int r2= getRegion(rc);
            if (r2>0&&r2!=region) return true;
         }
      }
      return false;
   }
   int countLandNeighbors(int r, int c){
      int N=MapHex.numNeighbors();
      int count=0;
      for (int i=0; i<N; i++){
         int rc[]= MapHex.getNeighborCoordinates(i,r,c);
         if (inBounds(rc)){
            if (grid[rc[0]][rc[1]].terrain==TerrainTypes.LAND)
               count++;
         }
      }
      return count;
   }
   boolean isPort(int rc[]){
      for (int i=0; i<numNeighbors; i++){
         int rc2[]= MapHex.getNeighborCoordinates(i,rc[0],rc[1]);
         if (inBounds(rc2)){
            if (getTerrain(rc2)==TerrainTypes.SEA){
               if (getRegion(rc2)==externalSea){
                  return true;
               }
            }
         }
      }
      return false;
   }

   // utilities for map generation

   // converts a vector of cells into a hashtable of cells
   Hashtable vect2hash(Vector v){
      Hashtable ht= new Hashtable();
      int N= v.size();
      for (int i=0; i<N; i++){
         int rc[]= (int [])v.elementAt(i);
         ht.put(rc[0]+","+rc[1],rc);
      }
      return ht;
   }
   // converts a hashtable of cells into a vector of cells
   Vector hash2vect(Hashtable ht){
      Vector V= new Vector();
      Object x[]= ht.values().toArray();
      for (int i=0; i<x.length; i++){
         V.add(x[i]);
      }
      return V;
   }
   TerrainMap getTerrainMap(int terrain){
      return new TerrainMap(terrain);
   }
   void markExternalSea(){
      TerrainMap tm= new TerrainMap(TerrainTypes.SEA);

      // find the largest sea region.
      int N= tm.regions.size();
      int maxSize= 0;
      int maxRegion= 0;
      for (int i=0; i<N; i++){
         Vector region= (Vector)tm.regions.elementAt(i);
         int size= region.size();
         if (size>maxSize){
            maxSize= size;
            maxRegion= i+1;
            seaCells= region;
         }
      }
      externalSea= maxRegion;
   }
   void growLandMasses(int thresh){
      int N= landCells.size();
      for (int i=0; i<N; i++){
         int rc[]= (int[])landCells.elementAt(i);
         int region= getRegion(rc);
         int x= countLandNeighbors(rc[0],rc[1]);
         if (x<thresh){
            // pick a random neighbor
            int rc2[];
            do {
               int neigh= SystemUtil.rint(6);
               rc2= MapHex.getNeighborCoordinates(neigh,rc[0],rc[1]);
            } while (!inBounds(rc2));

            // make it a LAND hex if it doesn't join regions
            if (getTerrain(rc2)==TerrainTypes.SEA){
               if (!connectsRegion(rc2[0],rc2[1],region)){
                  set(rc2,TerrainTypes.LAND,region);
                  landCells.addElement(rc2);          
               }
            }
         }
      }
   }
   void sprinklePieces(int playerList[],int numPiecesOfType[]){
      // copy grid[][] to a Map object, if this hasn't already been done.
      if (m==null) copyToMap();

      // build list of all unoccupied land cells
      Vector availableLandV= new Vector();
      int N= landCells.size();
      for (int i=0; i<N; i++){
         int rc[]= (int[])landCells.elementAt(i);
         if (m.getPieceAt(rc[0],rc[1])==null){
            availableLandV.add(rc);
         }
      }

      // grab list of all unoccupied sea cells
      Vector availableSeaV= seaCells;

      for (int i=0; i<playerList.length; i++){
         for (int j=0; j<numPiecesOfType.length; j++){
            for (int k=0; k<numPiecesOfType[j]; k++){
               // get new piece
               Piece p= new Piece(j,playerList[i],0,0);

               // by land or by sea?
               Vector V;
               if (p.canMoveOver(TerrainTypes.LAND)){
                  V= availableLandV;
               }
               else {
                  V= availableSeaV;
               }

               // choose random map cell from available ones
               N= V.size();
               if (N>0){
                  int l= SystemUtil.rint(N);
                  int rc[]= (int[])V.elementAt(l);

                  // remove it from the available list
                  V.removeElementAt(l);

                  // and put the new piece there
                  p.setRowAndColumn(rc[0],rc[1]);
                  m.attachPiece(p);
               }
            }
         }
      }
   }
   void sprinkleCities(int numCities, int minPorts){
      // copy grid[][] to a Map object, if this hasn't already been done.
      if (m==null) copyToMap();

      TerrainMap tm= new TerrainMap(TerrainTypes.LAND);

      markExternalSea();

      // for each land mass detected ...
      int N= tm.regions.size();
      for (int i=0; i<N; i++){
         Vector region= (Vector)tm.regions.elementAt(i);

         // get a list of all possible ports
         Vector possiblePorts= new Vector();
         int M= region.size();
         for (int j=0; j<M; j++){
            int rc[]= (int[])region.elementAt(j);
            if (isPort(rc)){
               possiblePorts.add(rc);
            }
         }

         Hashtable htRegion= vect2hash(region);

         // randomly select minPorts, if possible, from all possible ports
         // and put neutral cities there.
         int npp= possiblePorts.size();
         int np= Math.min(minPorts,npp);
         int ports[]= SystemUtil.NChooseK(npp,np);
         for (int j=0; j<ports.length; j++){
            // place a city at the given port cell
            int rc[]= (int[])possiblePorts.elementAt(ports[j]);
            m.attachPiece(new Piece(cityPieceType,neutralPlayerNumber,rc[0],rc[1]));

            // remove the port from the list of remaining land cells in this region
            htRegion.remove(rc[0]+","+rc[1]);
         }

         // randomly place additional cities on the land mass
         Object cells[]= htRegion.values().toArray();
         M= Math.min(cells.length,numCities);
         int cities[]= SystemUtil.NChooseK(cells.length,M);
         for (int j=0; j<cities.length; j++){
            int rc[]= (int [])cells[cities[j]];
            m.attachPiece(new Piece(cityPieceType,neutralPlayerNumber,rc[0],rc[1]));            
         }
      }
   }
   void copyToMap(){
      try {
         // create all sea Map
         m= new Map(null,numRows,numCols,numRows,numCols);

         // copy terrain values to it.
         for (int r=0; r<numRows; r++){
            for (int c=0; c<numCols; c++){
               m.setTerrain(r,c,getTerrain(r,c));
            }
         }
      }
      catch (BadRowAndColumnValuesException e){
         System.err.println(e);
         System.exit(1);
      }
   }

   // I/O

   public void write(PrintStream out){
      // copy grid[][] to a Map object, if this hasn't already been done.
      if (m==null) copyToMap();

      // write to file
      m.write(new OutputStreamWriter(out));
   }

   public static void main1(String args[]) throws Exception{
      String usage="java GEMpire.GenMap <num rows> <num cols> <num seeds> <num grows> <thresh>";
      if (args.length<5){
         System.err.println(usage);
         System.exit(1);
      }

      int numRows= Integer.parseInt(args[0]);
      int numCols= Integer.parseInt(args[1]);
      int numSeeds= Integer.parseInt(args[2]);
      int numGrows= Integer.parseInt(args[3]);
      int thresh= Integer.parseInt(args[4]);

      Globals.packageStaticInit(true); 
      GenMap gm= new GenMap(numRows,numCols,numSeeds,numGrows,thresh);
      gm.write(System.out);
      System.exit(0);
   }
   public static void main(String args[]) throws Exception{
      String usage="java GEMpire.GenMap <num rows> <num cols> <num seeds> <land ratio>"+
                   " <thresh> <quit delta> <num cities> <num ports>";
      if (args.length<6){
         System.err.println(usage);
         System.exit(1);
      }

      int numRows= Integer.parseInt(args[0]);
      int numCols= Integer.parseInt(args[1]);
      int numSeeds= Integer.parseInt(args[2]);
      double landRatio= new Double(args[3]).doubleValue();
      int thresh= Integer.parseInt(args[4]);
      double quitDelta= new Double(args[5]).doubleValue();
      int numCities= Integer.parseInt(args[6]);
      int numPorts= Integer.parseInt(args[7]);

      Globals.packageStaticInit(true); 
      GenMap gm= new GenMap(numRows,numCols,numSeeds,landRatio,thresh,quitDelta);
      gm.sprinkleCities(numCities,numPorts);

      int playerList[]={1,2,3};
      int N= 3;
      int numPiecesOfType[]={0,N,N,0,0,0,N,N,N,N,N,N};
      gm.sprinklePieces(playerList,numPiecesOfType);

      gm.write(System.out);
      System.exit(0);
   }
   public static void main3(String args[]) throws Exception{
      String usage="java GEMpire.GenMap <map file>";
      if (args.length<1){
         System.err.println(usage);
         System.exit(1);
      }

      String filename=args[0];

      Globals.packageStaticInit(true); 
      GenMap gm= new GenMap(filename);
      TerrainMap tm= gm.getTerrainMap(TerrainTypes.LAND);
      tm.write(System.out);
      System.exit(0);
   }
   public static void main4(String args[]) throws Exception{
      String usage="java GEMpire.GenMap <map file> <cities per land mass> <min ports per land mass>";
      if (args.length<3){
         System.err.println(usage);
         System.exit(1);
      }

      String filename=args[0];
      int numCities= Integer.parseInt(args[1]);
      int minPorts= Integer.parseInt(args[2]);

      Globals.packageStaticInit(true); 
      GenMap gm= new GenMap(filename);
      gm.sprinkleCities(numCities,minPorts);
      gm.write(System.out);
      System.exit(0);
   }
}




