//FILE: Map.java

/* GEMpire (Generic Empire for the Masses), a java package to support internet oriented 
 * empire-like strategy games and associated utilities.
 * Copyright (C) 1998 Christopher L. Marshall
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * My current email is christopherlmarshall@yahoo.com.  My home address is:
 * 10006 Leamoore Lane,
 * Vienna VA, 22181
 * 
 * It is my hope that this modest effort might assist others in producing Java
 * programs to implement a varied host of empire-like games.
 */

package GEMpire;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

/** Displays a hex map with scrollbars.  I did not use a scrollpane because every paint would result in
  * the whole map being drawn (albeit off screen) and I imagine this would be noticably slower than 
  * managing the scrollbars myself and only having the visible portion painted.
  */
public class Map extends Panel{
   // Default values for new instances.
      static int defaultNumDisplayRows= 4;
      static int defaultNumDisplayCols= 2;

   // Main instance variables for storing Map contents.

      /** this is the 2d array of cells containing terrain and piece info for the map. */
      private MapHex grid[][];
      /** dimensions of grid[][] */
      private int numRows, numCols;

   // location and size of the visible portion of the map.

      private int numDisplayRows, numDisplayCols;
      private int leftDisplayColumn=0;
      private int topDisplayRow=0;

   // Utility objects.
      /** Translates mouse clicks into row and column 'events'. */
      private HexCoordinateConverter hcc;

   // Interface objects (to handle mouse events, key events, piece movements, ...) and
   // their accessor methods.

      /** Handles movement key events */
      private MapKeyAdapter mka=null;

      /** Handles non-movement key events */
      private KeyReceiver kr= null;

      /** PieceMover */
      private PieceMover pm;

      /** gets the PieceMover */
      public PieceMover getPieceMover(){return pm;}

      /** sets the KeyReceiver */
      public void setKeyReceiver(KeyReceiver kr){
         this.kr= kr;
         if (mka!=null){
            mka.setKeyReceiver(kr);
         }
      }

      /** sets the PieceMover.  Can be set after creation to a different version 
        * depending how the Map instance is being used (in an editor, in a client, 
        * in a server, ...)
        */
      public void setPieceMover(PieceMover pm){
         this.pm= pm;
         if (mka!=null) mka.setPieceMover(pm);
      }

      /** this listener determines what happens when a hex is clicked with the mouse.  The default
        * action is to activate the piece occupying the hex, if any.  This can be changed by calling
        * Map.setRowAndColumnSelectedListener(), and would be if the map were used in a MapEditor,
        * for example.  The default is appropriate for a client program using a Map.
        */
      private RowAndColumnSelectedListener racsl;

      /** sets the RowAndColumnSelectedListener() that is called when a location on the
        * is selected.
        */
      public void setRowAndColumnSelectedListener(RowAndColumnSelectedListener racsl){
         this.racsl= racsl;
      }

      /** This listener is called when a command is issued to a piece (like a production change).
        * The EmpireClient is an example.  It needs to send a message to the EmpireServer when the
        * player changes production.
        */
      private PieceCommandListener pcListener;
      public void setPieceCommandListener(PieceCommandListener pcl){pcListener= pcl;}

   // Constants, fields, and methods for piece select mode implementation.  
      /* When the map is in 
       * PRODUCTION_SELECT_MODE, mouse clicks on fixed pieces brings up a production dialog.  When
       * the map is in PIECE_SELECT_MODE, mouse clicks on fixed pieces activates pieces parked there.
       */
      public static final int PRODUCTION_SELECT_MODE= 0;
      public static final int PIECE_SELECT_MODE     = 1;
      private int selectMode= PIECE_SELECT_MODE;
      public void setSelectMode(int i){selectMode= i;}
      private int getSelectMode(){return selectMode;}

   // Components
      /** The MapCanvas is a trival extension of Canvas which receives paint() calls and 
        * calls the Map paint routine.  This allows me to keep the drawing code in Map,
        * where is belongs, even though Map is really a container (a Panel to be exact),
        * and containers don't normally do custom drawing. 
        */
      private MapCanvas mc;

      /** Scrolls the visible portion of the Map through the underlying grid
        * of cells.
        */
      private Scrollbar h,v;
      /** Horizontal RulerBar which displays column numbers across the top of the map. */
      private RulerBar hrb;
      private RulerBar vrb;

   // main ivar accessor methods
      int getNumRows(){return numRows;}
      int getNumCols(){return numCols;}
   // Terrain and other map-cell manipulation accessor methods.
      public int  getTerrain(int row, int column){return grid[row][column].getTerrain();}
      public void setTerrain(int row, int column, int terrain){grid[row][column].setTerrain(terrain);}
      public void setAllTerrain(int terrain){
         for (int r=0; r<numRows; r++){
            for (int c=0; c<numCols; c++){
               grid[r][c].setTerrain(terrain);
            }
         }
      }

      public void markSurroundingCellsAsExploredBy(int row, int column, int player){
         grid[row][column].markAsExploredBy(player);
         for (int i=0; i<MapHex.numNeighbors(); i++){
            int rc[]= MapHex.getNeighborCoordinates(i,row,column);
            if (inBounds(rc[0],rc[1])){
               grid[rc[0]][rc[1]].markAsExploredBy(player);
            }
         }
      }

      public boolean[] getVisibleStatus(int row, int col, boolean players[]){
         return grid[row][col].getVisibleStatus(players);
      }

      /** If p at row,col can see any neighbor, set visibility true.
        */
      public void markVisibilityAround(int row, int col){
         Piece p= getPieceAt(row,col);
         if (p==null) Globals.exit(
            "Map.markVisibilityAround("+row+","+col+"): cell is empty!  This shouldn't happen!"
         );

         grid[row][col].setVisibilityTo(p.getOwner(),true);
         for (int i=0; i<MapHex.numNeighbors(); i++){
            int rc[]= MapHex.getNeighborCoordinates(i,row,col);
            if (inBounds(rc[0],rc[1])){
               if (p.canSee(getPieceAt(rc[0],rc[1])))
                  grid[rc[0]][rc[1]].setVisibilityTo(p.getOwner(),true);
            }
         }
            
      }

      /** can player see row,col? */
      private void determineVisibilityOfCellToPlayer(int row, int col, int player){
         MapHex mh= grid[row][col];
         mh.setVisibilityTo(player,false);

         Piece p= getPieceAt(row,col);
         if (p==null) return; // empty cells are invisible, for all practical purposes.
         if (p.getOwner()==player){
            mh.setVisibilityTo(player,true);
            return;
         }

         for (int i=0; i<MapHex.numNeighbors(); i++){
            int rc[]= MapHex.getNeighborCoordinates(i,row,col);
            if (inBounds(rc[0],rc[1])){
               Piece p2= getPieceAt(rc[0],rc[1]);
               if (p2!=null) if (p2.getOwner()==player) if (p2.canSee(p)){
                  mh.setVisibilityTo(player,true);
                  return;
               }
            }
         }
      }

      /** Can player see cells around row,col? */
      public void determineVisibilityOfCellsSurrounding(int row, int col, int player){
         for (int i=0; i<MapHex.numNeighbors(); i++){
            int rc[]= MapHex.getNeighborCoordinates(i,row,col);
            if (inBounds(rc[0],rc[1])){
               determineVisibilityOfCellToPlayer(rc[0],rc[1],player);
            }
         }
      }

      /** Can neighbors of row,col see it? */
      public void markVisibilityOfCellToNeighbors(int row, int col){
         Piece p= getPieceAt(row,col);
         if (p==null) return;
         
         int player= p.getOwner();
         MapHex mh= grid[row][col];
         mh.clearVisibilityFlags();
         mh.setVisibilityTo(player,true);

         for (int i=0; i<MapHex.numNeighbors(); i++){
            int rc[]= MapHex.getNeighborCoordinates(i,row,col);
            if (inBounds(rc[0],rc[1])){
               Piece p2= getPieceAt(rc[0],rc[1]);
               if (p2!=null) if (p2.canSee(p)){
                  mh.setVisibilityTo(p2.getOwner(),true);
               }
            }
         }
      }

      public void updateLastKnownOwners(Piece p){
         grid[p.getRow()][p.getColumn()].updateLastKnownOwners();
      }

   // Methods and instance variables for manipulating pieces on the map.
      private boolean[] players;
      private Vector pieces= new Vector();
      public final boolean[] getPlayers(){
         boolean tmp[]= new boolean[players.length];
         for (int i=0; i<players.length; i++)
            tmp[i]= players[i];
         return tmp;
      }
      public boolean canAttack(Piece a, Piece b){
         if (b.isFixed()||a.canMoveOver(getTerrain(b.getRow(),b.getColumn()))) return true;
         return false;
      }
      public final Piece selectPieceAt(int row, int column){return grid[row][column].selectPiece();}
      public final Piece getPieceAt(int row, int column){return grid[row][column].getPiece();}
      public final Piece getPieceAt(int id, int row, int column){return grid[row][column].getPiece(id);}
      public final Vector getPiecesAt(int row, int column){return grid[row][column].getPieces();}
      public final boolean hasFixedPiece(int row, int column){
         Piece p= getPieceAt(row,column);
         if (p==null) return false;
         if (p.isFixed()) return true;
         return false;
      }
      public final int getOwner(int row, int column){
         return grid[row][column].getOwner();
      }

      /**  attaches the piece p to the cell of the map at its row and column indexes. */
      public boolean attachPiece(Piece p){
         MapHex mh= grid[p.getRow()][p.getColumn()];
         boolean tmp= mh.attachPiece(p);
         if (tmp){
            markSurroundingCellsAsExploredBy(p.getRow(),p.getColumn(),p.getOwner());
            mh.setVisibilityTo(p.getOwner(),true);
            return true;
         }
         return false;
      }

      /**  detaches the piece p from the cell of the map at its row and column indexes. */
      public boolean detachPiece(Piece p){
         try {
            if (p==null) return false;
            return grid[p.getRow()][p.getColumn()].detachPiece(p);
         }
         catch (ArrayIndexOutOfBoundsException e){ 
            return false;
         }
      }

      /** this is a subclass of thread which controls the blinking of the piece clicked on by the user.
        */ 
      private PieceBlinker blinker;

      /** returns Piece currently selected by user for movement or manipulation. */
      public Piece getActivePiece(){if (blinker==null) return null; else return blinker.getPiece();}

      /**  de-selects a piece;  the piece no longer responds to the arrow keys.
        */
      public void deactivatePiece(){
         if (blinker!=null){
            blinker.stop();
            try {blinker.join();} catch (InterruptedException e){Globals.exit("could not join blinker"+e.toString());}
            int rc[]= blinker.getRowAndColumn();
            drawCell(rc[0],rc[1]);
         }
         blinker= null;
      }

      /**  causes a piece on the map at (row,col) to blink and respond to direction keys.
        */
      public void activatePiece(int row, int col){
         if (isVisible()){
            deactivatePiece();
            if (getSelectMode()==PIECE_SELECT_MODE) {
               Piece p= selectPieceAt(row,col);
               if (p!=null){
                  pcListener.pieceCommand(p,PieceCommandListener.PIECE_ACTIVATE);
                  blinker= new PieceBlinker(this,p,500,500);
                  blinker.start();
               }
            }
            else if (getSelectMode()==PRODUCTION_SELECT_MODE){
               Piece p= getPieceAt(row,col);
               if (p!=null){
                  p.displayProductionDialog(f);
                  pcListener.pieceCommand(p,PieceCommandListener.PRODUCTION_MENU_ACTIVATE);
               }
            }
            else {
               Globals.exit("Map.activatePiece("+row+","+col+"): invalid select mode: "+getSelectMode());
            }
         }
      }

      public boolean activatePiece(Piece p){return activatePiece(p.getId(),p.getRow(),p.getColumn());}
      public boolean activatePiece(int id, int row, int col){
         if (isVisible()){
            deactivatePiece();
            Piece p= getPieceAt(id,row,col);
            if (p!=null){
               blinker= new PieceBlinker(this,p,500,500);
               blinker.start();
               return true;
            }
         }
         return false;
      }

      public Vector implementNewTurn(){
         Vector deadPieces= new Vector();
         for (int r=0; r<numRows; r++){
            for (int c=0; c<numCols; c++){
               Piece p;
               if ((p=getPieceAt(r,c))!=null){
                  p.endTurn();
                  if (p.isAlive()){
                     Piece p2= p.refresh();
                     if (p2!=null){
                        // This condition occurs when a piece has built another piece, like an infantry building
                        // an airbase.  The old piece is replaced by the newly built piece.
                        if (!detachPiece(p)){
                           System.err.println(
                              "Map.implementNewTurn(): could not detach piece, "+p+
                              "This shouldn't happen!"
                           );
                        }
                        if (!attachPiece(p2)){
                           System.err.println(
                              "Map.implementNewTurn(): could not attach newly built piece, "+p2+
                              "This shouldn't happen!"
                           );
                        }
                     }
                  }
                  else {
                     detachPiece(p);
                     determineVisibilityOfCellsSurrounding(p.getRow(),p.getColumn(),p.getOwner());
                     deadPieces.addElement(p);
                  }
               }
            }
         }
         return deadPieces;
      }

      public boolean[] findAllPlayers(){
         for (int i=0; i<players.length; i++) players[i]= false;
         for (int r=0; r<numRows; r++){
            for (int c=0; c<numCols; c++){
               int owner= getOwner(r,c);
               if (owner>=0) players[owner]= true;
            }
         }
//for (int i=0; i<players.length; i++) if (players[i]) System.out.println("player "+i+" is still alive");
         return players;
      } 

   // Accessor methods and objects for general Map properties and modes.

      public void setDisplaySize(int numDisplayRows, int numDisplayCols, Frame f){
         this.numDisplayRows= Math.min(numDisplayRows,numRows);
         this.numDisplayCols= Math.min(numDisplayCols,numCols);
         if (f!=null) allocComponents(f);
      }

      public boolean withinBounds(int row, int column){
         if (row>=0&&row<numRows&&column>=0&&column<numCols) return true;
         return false;
      }
      public Dimension getRowAndColumnDisplayDimensions(){return new Dimension(numDisplayCols,numDisplayRows);}
      public Dimension getRowAndColumnDimensions(){return new Dimension(numCols,numRows);}
      final public int screenColumn(int column){return column-leftDisplayColumn;}
      final public int screenRow(int row){return row-topDisplayRow;}
      final public int getAbsRow(int row){return row+topDisplayRow;}
      final public int getAbsColumn(int column){return column+leftDisplayColumn;}
      boolean inBounds(int row, int col){
         return (row>=0)&&(row<numRows)&&(col>=0)&&(col<numCols);
      }

   // Drawing methods.

      public void drawBlack(int row, int column){
         Graphics g= mc.getGraphics();
         if (g!=null){
            grid[row][column].drawBlack(g,screenRow(row),screenColumn(column),mc);
            g.dispose();
         }
      }

      public void drawPattern(int row, int column, int pattern){
         Graphics g= mc.getGraphics();
         if (g!=null){
            grid[row][column].drawPattern(g,screenRow(row),screenColumn(column),pattern,mc);
            g.dispose();
         }
      }

      public void drawPiece(Piece p){
         Graphics g= mc.getGraphics();
         int row= p.getRow();
         int column= p.getColumn();
         if (g!=null){
            grid[row][column].drawTerrain(g,screenRow(row),screenColumn(column),mc);
            grid[row][column].drawPiece(g,p,screenRow(row),screenColumn(column),mc);
            g.dispose();
         }
      }

      public synchronized void drawCell(int row, int column){
         Graphics g= mc.getGraphics();
         if (g!=null){
            grid[row][column].draw(g,screenRow(row),screenColumn(column),mc);
            g.dispose();
         } 
      }

      public synchronized void drawCellTerrain(int row, int column){
         Graphics g= mc.getGraphics();
         if (g!=null){
            grid[row][column].drawTerrain(g,screenRow(row),screenColumn(column),mc);
            g.dispose();
         }
      }

      /** Used to detect changes in display size. */
      private Dimension curSize;

      /** paints central map. */
      protected void paintToCanvas(Graphics g, Dimension d){
         if (!curSize.equals(d)){
            Dimension tmp= MapHex.convertPixelSizeToRowColSize(d);
            setDisplaySize(tmp.height,tmp.width,null);
            curSize= d;
            h.setValues(leftDisplayColumn,numDisplayCols,0,numCols-1);
            v.setValues(topDisplayRow,numDisplayRows/2,0,numRows/2-1);
            hrb.setRange(leftDisplayColumn,numDisplayCols);
            vrb.setRange(topDisplayRow,numDisplayRows);
         }

         for (int r=0; r<numDisplayRows; r++){
            if (r%2==0){
               int r_2= r/2;
               for (int c=0; c<numDisplayCols; c++){
                  grid[r+topDisplayRow][c+leftDisplayColumn].drawAsEvenRow(g,r_2,c,mc);
               }
            }
            else {
               int r_1_2= (r-1)/2;
               for (int c=0; c<numDisplayCols; c++){
                  grid[r+topDisplayRow][c+leftDisplayColumn].drawAsOddRow(g,r_1_2,c,mc);
               }
            }
         }
      }

   // Allocation and Initialization.

   /** Make sure the number of rows and columns for the map and the display are all even and perform 
    *  other sanity checks
     */
   void checkRowAndColumnValues() throws BadRowAndColumnValuesException{
      if ((numRows%2!=0)||(numCols%2!=0)||(numDisplayRows%2!=0)||(numDisplayCols%2!=0)) 
         throw new BadRowAndColumnValuesException(
            "numRows, numCols, numDisplayRows, and numDisplayCols must all be even"
         );
      if (numRows<numDisplayRows) numDisplayRows= numRows;
      if (numCols<numDisplayCols) numDisplayCols= numCols;
   }

   Frame f;
   /** Create and initialize scrollbars and required listeners.
     */
   void allocComponents(Frame f){
      this.f= f;

      // allocComponents may be called more than once on the same instance so we need to
      // remove any previous components before adding new ones.
      removeAll();

      // create center display canvas.
      mc= new MapCanvas(this, numDisplayRows, numDisplayCols);
      mc.setBackground(Color.black);
      add(mc,"Center");

      // create ruler bars.
      try {
         vrb= new RulerBar(
            null,
            0,
            MapHex.getRowSpacing(),
            topDisplayRow,
            topDisplayRow+numDisplayRows-1,
            numRows-1,
            false
         );
         hrb= new RulerBar(
            null,
            vrb.getPreferredSize().width,
            MapHex.getColumnSpacing(),
            leftDisplayColumn,
            leftDisplayColumn+numDisplayCols-1,
            numCols-1,
            true
         );
      }
      catch (RulerBarSpacingTooSmallException e){
         Globals.exit(
            "Incorrect spacing of RulerBar in Map.allocComponents():  "+
            "This shouldn't happen."+e
         );
      }

      add(vrb,"West");
      add(hrb,"North");

      hcc= new HexCoordinateConverter(MapHex.getRowWidths());

      // Create default RowAndColumnSelectedListener to activate any piece found at 
      // (row,column)
      racsl= new RowAndColumnSelectedListener(){
         public void selected(int row, int column){
            activatePiece(row,column);
         }
      };

      // add mouse listener to map canvas
      mc.addMouseListener(new MouseAdapter(){
         public void mousePressed(MouseEvent e){
            hcc.convert(e.getX(),e.getY());
            racsl.selected(getAbsRow(hcc.getRow()),getAbsColumn(hcc.getColumn()));
         }
      });

      // Initialize KeyListener and key event processing mechanism.  Essentially, the MapKeyAdapter receives
      // KeyEvents from the MapCanvas or the top level frame (jdk1.1 and jdk1.1.5 differ on this so I had to
      // do it both ways to maximize compatibility).  For movement commands (the direction keys), the MapKeyAdapter
      // calls the PieceMover directly.  For all other key events, MapKeyAdapter passes the event on to the
      // KeyReceiver, which in one case is the EmpireServerConnection (the connection to the server that the 
      // EmpireClient has, and where most of the work associated with the EmpireClient actually 
      // gets done), but doesn't have to be, hence the ability to arbitrarly assign where the key events go.

      if (mka!=null) mka.removeFromFrame();
      if (pm==null) pm= new PieceMover(this);
      mka= new MapKeyAdapter(this,f,pm); // mka will handle key events from the canvas or frame
      mka.setKeyReceiver(kr);            // kr will process all non movement key events passed in from mka
      mc.addKeyListener(mka);

      pcListener= new PieceCommandAdapter(){
         public void PieceCommand(Piece p){}
      };

      // properly configures scrollbars given numRows, numCols, numDisplayRows, and
      // numDisplayCols.
      allocateScrollbars();
   }

   /** creates scrollbars to control visible portion of grid according to values in
     * numRows, numCols, numDisplayRows, and numDisplayCols.
     */
   void allocateScrollbars(){
      curSize= MapHex.convertRowColSizeToPixelSize(
         new Dimension(numDisplayCols, numDisplayRows)
      );

      // create bottom scrollbar
      if (h!=null) remove(h);
      h= new Scrollbar(Scrollbar.HORIZONTAL,0,numDisplayCols,0,numCols-1);
      h.addAdjustmentListener(new AdjustmentListener(){
         public void adjustmentValueChanged(AdjustmentEvent e){
            int tmp= e.getValue();
            if (tmp<=(numCols-numDisplayCols)) 
               leftDisplayColumn= tmp;
            hrb.setRange(leftDisplayColumn,numDisplayCols);
            mc.repaint();
         }
      });
      add(h,"South");

      // create right scrollbar
      if (v!=null) remove(v);
      v= new Scrollbar(Scrollbar.VERTICAL,0,numDisplayRows/2,0,numRows/2-1);
      v.addAdjustmentListener(new AdjustmentListener(){
         public void adjustmentValueChanged(AdjustmentEvent e){
            int tmp= e.getValue();
            if (tmp<=(numRows-numDisplayRows)/2) 
               topDisplayRow= 2*tmp;
            vrb.setRange(topDisplayRow,numDisplayRows);
            mc.repaint();
         }
      });
      add(v,"East");
   }

   public void removeFromFrame(){
      if (mka!=null){
         mka.removeFromFrame();
      }
   }

   /**  Allocate square array of cells
     */
   void hexAlloc(){
      grid= new MapHex[numRows][numCols];
      for (int i=0; i<numRows; i++) 
         for (int j=0; j<numCols; j++) 
            grid[i][j]= new MapHex();
   }

   void resizeGrid(int nr, int nc){
      //Allocate new array and copy from old
      MapHex tmp[][]= new MapHex[nr][nc];
      int r,c,numRowsToCopy=Math.min(numRows,nr),numColsToCopy=Math.min(numCols,nc);
      for (r=0; r<numRowsToCopy; r++){
         for (c=0; c<numColsToCopy; c++){
            tmp[r][c]= (MapHex)(grid[r][c].clone());
         }
         for (c=numColsToCopy; c<nc; c++){
            tmp[r][c]= new MapHex();
         }
      }
      for (r=numRowsToCopy; r<nr; r++){
         for (c=0; c<nc; c++){
            tmp[r][c]= new MapHex();
         }
      }
      grid= tmp;
      numRows= nr;
      numCols= nc;
      curSize= null;
      allocComponents(f);
      invalidate();
   }

    /** creates all-sea map with the specified size and display size.
     */
   public Map(Frame f, int numRows1, int numCols1, int numDisplayRows1, int numDisplayCols1) 
      throws BadRowAndColumnValuesException
   {
      super(new BorderLayout());

      numCols= numCols1;
      numRows= numRows1;
      numDisplayRows= numDisplayRows1;
      numDisplayCols= numDisplayCols1;

      checkRowAndColumnValues();
      if (f!=null) allocComponents(f);
      hexAlloc();
   }

   /**  Reads map info from text file, and sets dimensions of on-screen display of map.
     */
   public Map(Frame f, Reader r) 
      throws IOException, BadRowAndColumnValuesException
   {
      super(new BorderLayout());

      // Read number of rows and columns from file.
      BufferedReader in= new BufferedReader(r);
      String line= in.readLine();
      numRows= Integer.parseInt(line.trim());
      line= in.readLine();
      numCols= Integer.parseInt(line.trim());

      numDisplayRows= defaultNumDisplayRows;
      numDisplayCols= defaultNumDisplayCols;
      checkRowAndColumnValues();

      if (f!=null) allocComponents(f);
      hexAlloc();

      DelimitedLineReader dlr= new DelimitedLineReader(',');
      for (int i=0; i<numRows; i++){
         line= in.readLine();
         dlr.setLine(line);
         for (int j=0; j<numCols; j++){
            grid[i][j].setTerrain(dlr.getNextInt());
         }
      }

      // Read descriptions of any pieces on the map and attach. Note Players.  Create list of pieces.
      players= new boolean[Piece.numPossiblePlayers()];
      for (int i=0; i<players.length; i++) players[i]= false;
      int numPieces;
      line= in.readLine();
      numPieces= Integer.parseInt(line.trim());
      Piece p;
      for (int i=0; i<numPieces; i++){
         line= in.readLine();
         if (!attachPiece(p= new Piece(line)))
            Globals.exit(
               "Map.Map(): could not attach piece "+p+" read from file.  "+
               "This shouldn't happen!"
            );
         pieces.addElement(p);
         if (!p.isNeutral()){
            players[p.getOwner()]= true;
         }
         markSurroundingCellsAsExploredBy(p.getRow(),p.getColumn(),p.getOwner());
      }

      Enumeration tmp= pieces.elements();
      while (tmp.hasMoreElements()){
         p= (Piece)tmp.nextElement();
         markVisibilityAround(p.getRow(),p.getColumn());
      }

      // Read player explored patterns, if any.
      int numPatterns= 0;
      line= in.readLine().trim();
      numPatterns= Integer.parseInt(line);
      for (int i=0; i<numPatterns; i++){
         readExploredPattern(in);
      }
   }

   public String writeCellUpdateTo(int row, int col, int player, boolean assumeVisible){
      if (inBounds(row,col)) if (grid[row][col].isExploredBy(player)||player==0) {
         StringWriter sw= new StringWriter();
         PrintWriter out= new PrintWriter(sw);
         out.println(row);
         out.println(col);
         grid[row][col].writeCellUpdateTo(out,player,assumeVisible);
         out.flush();
         return sw.toString();
      }
      return null;
   }

   public boolean readCellUpdate(String update){
      try {
         BufferedReader in= new BufferedReader(new StringReader(update));
         int row= Integer.parseInt(in.readLine().trim());
         int col= Integer.parseInt(in.readLine().trim());
         if (!inBounds(row,col)){
            Globals.reportError(
               "Map.readCellUpdate(): (row,column)= "+row+","+col+
               " is out of bounds of the defined Map area."
            );
            return false;
         }
         boolean tmp= grid[row][col].readCellUpdate(in);
         drawCell(row,col);
         return tmp;
      }
      catch (IOException e){
         return false;
      }
      catch (NullPointerException e){
         return false;
      }
   }

// a shameless debugging hack
/*
void printAllPieces(){
   for (int r=0; r<numRows; r++){
      for (int c=0; c<numCols; c++){
         Piece p= getPieceAt(r,c);
         if (p!=null)
            System.err.println(""+p);
      }
   }
}
*/

   public void write(Writer out){
      write(out,0);
   }
   public void write(Writer out, int player){

//System.err.println("Map.write(): printing all pieces");
//printAllPieces();
      int i,j;
      Vector v_pieces= new Vector(100,100);
      PrintWriter bout= new PrintWriter(out);

      // Write number of rows and columns.
      bout.println(numRows);
      bout.println(numCols);

      // Write terrain value for all rows and columns
      Piece p;
      for (i=0; i<numRows; i++){
         for (j=0; j<numCols; j++){ 
            // Write terrain value.
            if (j<numCols-1)
               bout.print(grid[i][j].getTerrain(player)+","); 
            else
               bout.println(grid[i][j].getTerrain(player)); 

            if (grid[i][j].isExploredBy(player))
               if (grid[i][j].isVisibleTo(player)||hasFixedPiece(i,j)){
                  // Copy any pieces.
                  if ((p=getPieceAt(i,j))!=null) v_pieces.addElement(p);
               }
         }
      }

      // Write list of pieces attached to Map, keep track of players with pieces on map.
      boolean playersOnMap[]= new boolean[Piece.numPossiblePlayers()];
      for (i=0; i<playersOnMap.length-1; i++)
         playersOnMap[i]= false;
      bout.println(v_pieces.size());
      for (i=0; i<v_pieces.size(); i++){
         p= (Piece)v_pieces.elementAt(i);
         p.write(bout,player);
         int owner= p.getOwner();
         if (owner!=-1){
            playersOnMap[owner]= true;
         }
      }
//System.err.println("v_pieces.size()= "+v_pieces.size());

      // don't write exploration patterns if this map is being written for a specific player to see.
      if (player==0){
         // Count number of players with at least one piece on the map and write to file.
         int numPlayers= 0;
         for (i=0; i<playersOnMap.length; i++)
            if (playersOnMap[i]) numPlayers++;
         bout.println(numPlayers);

         // Write the exploration pattern for each player to file.
         for (i=0; i<playersOnMap.length; i++){
            if (playersOnMap[i])
               writeExploredPatternForPlayer(bout,i);
//               bout.println("exploration pattern for player "+i+" here.");
         }
      }
      else bout.println("0");

      bout.println();
      bout.flush();
   }

   /** This method writes a simple run length encoded description of the regions the given player has
     * and hasn't explored.
     * 
     * output format (always 2 lines):
     * <player number>
     * <flag> <section length> ... <flag> <section length>
     * 
     * example:
     * 2
     * E 10 U 15 E 10 ... U 5
     *
     * This says the for player 2, the first 10 cells in the map are explored, the next 15 are unexplored,
     * the next 10 are explored, and the last 5 are unexplored.  Cells are ordered by row then by column, 
     * with cell (0,0) as the first cell (in other words, cells are read off starting in the upper left
     * hand corner and then right to left and top to bottom, like reading words on a page in English).  
     * Note that the ordering is with respect to cell row and column numbers regardless of how the 
     * cells actually lay out in a hex grid.
     */
   private void writeExploredPatternForPlayer(PrintWriter out, int player){
      boolean explored= false;
      out.println(player);
      if (grid[0][0].isExploredBy(player)){
         out.print("E ");
         explored= true;
      }
      else out.print("U ");

      int currentCount= 0;
      for (int r=0; r<numRows; r++){
         for (int c=0; c<numCols; c++){
            if (grid[r][c].isExploredBy(player)){
               if (explored){
                  currentCount++;
               }
               else {
                  out.print(currentCount+" E ");
                  currentCount= 1;
                  explored= true;
               }
            }
            else {
               if (explored){
                  out.print(currentCount+" U ");
                  currentCount= 1;
                  explored= false;
               }
               else {
                  currentCount++;
               }
            }
         }
      }
      out.println(currentCount);
   }

   private boolean getNextExploredSection(
      DelimitedLineReader dlr, 
      char flag[], 
      int length[]
   ) throws IOException
   {
      String field;

      // get flag character.
      field=dlr.getNextNonBlankField();
      if (field==null) return false;
      flag[0]= field.trim().charAt(0);
      
      // get count
      field=dlr.getNextNonBlankField();
      if (field==null) return false;
      length[0]= Integer.parseInt(field.trim());

      return true;
   }

   private void readExploredPattern(BufferedReader in) throws IOException{
      boolean explored= false;
      int player= Integer.parseInt(in.readLine().trim());

      DelimitedLineReader dlr= new DelimitedLineReader(' ');
      dlr.setLine(in.readLine());
 
      int rcCount= 0;
      int r,c,j,k;
      char flag[]=  new char[1];
      int length[]= new int[1];
      while (getNextExploredSection(dlr,flag,length)){
         if (flag[0]=='E'){
            for (j=0; j<length[0]; j++){
               k= rcCount+j;
               r= k/numCols;
               c= k%numCols;
               grid[r][c].markAsExploredBy(player);
            }
         }
         else if(flag[0]!='U') throw new IOException("Bad explored block flag: "+flag[0]);
         rcCount+= length[0];
      }
   }

   void debugReportOn(PrintWriter out, int row, int col, boolean players[]){
      for (int i=0; i<players.length; i++){
         if (players[i]){
            out.print(""+i+"\t");
         }
      }
      out.println();
      for (int i=0; i<players.length; i++){
         if (players[i]){
            if (grid[row][col].isVisibleTo(i)) 
               out.print("1\t");
            else out.print("0\t");
         }
      }
      out.println("\tvisibility");
      for (int i=0; i<players.length; i++){
         if (players[i]){
            if (grid[row][col].isExploredBy(i)) 
               out.print("1\t");
            else out.print("0\t");
         }
      }
      out.println("\texplored status");
      out.println(writeCellUpdateTo(row,col,-1,true));
   }

   void debugReportOn1(PrintWriter out, int row, int col, boolean players[]){
      out.println(getPieceAt(row,col));
   }
}




