//FILE: MapHex.java

/* GEMpire (Generic Empire for the Masses), a java package to support internet oriented 
 * empire-like strategy games and associated utilities.
 * Copyright (C) 1998 Christopher L. Marshall
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * My current email is christopherlmarshall@yahoo.com.  My home address is:
 * 10006 Leamoore Lane,
 * Vienna VA, 22181
 * 
 * It is my hope that this modest effort might assist others in producing Java
 * programs to implement a varied host of empire-like games.
 */

package GEMpire;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.Vector;
import java.io.*;

class MapHex {
   public static final int RED_PATTERN= 0;

   /** terrain images */
   private static Image terrainImages[]= new Image[TerrainTypes.numTypes];
   /** land hex image */
//   private static Image land;
   /** sea hex image */
//   private static Image sea;
   /** Misc images. */
   private static Image black;
   private static Image pattern[];

   /** width of enclosing rectangle */
   private static int rectWidth;
   /** height of enclosing rectangle */
   private static int rectHeight;
   /** number of pixels between two columns in the same row */
   private static int columnSpacing;
   /** x coord of upper left corner of row 1 */
   private static int oddRowOffsetX;
   /** y coord of upper left corner of row 1 */
   private static int oddRowOffsetY;
   /** the width of the rows of the solid hexagon from top to middle */
   private static int rowWidths[]={12,14,14,16,16,18,18,20,20,22};

   // Instance variables and clone() method.

      private byte terrainType= (byte)TerrainTypes.SEA;

      /** Used to hold explored and visible status for all players. */
      private int attributes= 0;

      /** If the hex is occupied by a city or a piece, it is considered a 'detailed hex'
        * and requires more information.  As most hexes in the map will tend to be empty,
        * this arrangement saves space.
        */
      DetailedHex dh= null;

      public Object clone(){
         MapHex tmp= new MapHex();
         tmp.terrainType= terrainType;
         tmp.attributes= attributes;
         if (dh==null) tmp.dh= null;
         else tmp.dh= (DetailedHex)dh.clone();
         return tmp;
      }

   public void writeCellUpdateTo(PrintWriter out, int player, boolean assumeVisible){
      out.println(getTerrain());
      if (dh==null){
         out.println("H");
      }
      else {
         dh.writeCellUpdateTo(out, isVisibleTo(player), player, assumeVisible);
      }
   }
   public boolean readCellUpdate(BufferedReader in){
      try {
         // read terrain.
         int tmp= Integer.parseInt(in.readLine());
         if (!TerrainTypes.isValid(tmp)){
            Globals.reportError("MapHex.readCellUpdate(): bad terrain type read: "+tmp);
            return false;
         }
         setTerrain(tmp);

         // read indication of detailed hex presence.
         String line= in.readLine().trim();
         if (line.equals("D")){
           dh= new DetailedHex(in);
         }
         else if (line.equals("H")){
            dh= null;
         }
         else {
            Globals.reportError("MapHex.readCellUpdate(): improper cell type in stream");
            return false;
         }
      }
      catch (IOException e){
         Globals.reportError(e.toString());
         return false;
      }
      catch (NullPointerException e){
         Globals.reportError(e.toString());
         return false;
      }

      return false;
   }

   public final int getTerrain(){return (int)terrainType;}
   public final int getTerrain(int player){
      if (!isExploredBy(player))
         return TerrainTypes.UNEXPLORED;
      return (int)terrainType;
   }
   public void setTerrain(int t){terrainType= (byte)t;}

   static boolean initialized= false;
   public static boolean initialized(){
      return initialized;
   }
   public boolean empty(){
      if (dh==null) return true;
      return dh.empty();
   }
   public Piece selectPiece(){if (dh==null) return null; else return dh.selectPiece();}
   public Piece getPiece(){if (dh==null) return null; else return dh.getPiece();}
   public Piece getPiece(int id){if (dh==null) return null; else return dh.getPiece(id);}
   public Vector getPieces(){if (dh==null) return new Vector(); else return dh.getPieces();}

   /** This initializes all static fields of MapHex */
   public static void loadImages(boolean graphicsEnabled) 
   {
      if (!initialized()){
         if (graphicsEnabled){
            int outlineColor= Color.white.getRGB();

            black= GraphicsUtil.makeSolidHex(
               Color.black.darker().getRGB(),
               rowWidths,
               outlineColor
            );
            terrainImages[TerrainTypes.LAND]= GraphicsUtil.makeSolidHex(
               Color.green.darker().getRGB(),
               rowWidths,
               outlineColor
            );
            terrainImages[TerrainTypes.SEA]= GraphicsUtil.makeSolidHex(
               Color.blue.getRGB(),
               rowWidths,
               outlineColor
            );
            terrainImages[TerrainTypes.UNEXPLORED]= GraphicsUtil.makeSolidHex(
               Color.black.getRGB(),
               rowWidths,
               outlineColor
            );
            Frame f= new Frame();
            for (int i=0; i<TerrainTypes.numTypes; i++){
               String filename= TerrainTypes.filenames[i];
               Image im= null;
               if (filename!=null){
                  try {
                     if ((im=GraphicsUtil.getImage(filename))==null)
                         Globals.exit(
                           "MapHex.loadImages(): could not load image from "+filename
                        );
                     terrainImages[i]= GraphicsUtil.changeColor(im,Color.black.getRGB(),0,f);
                  }
                  catch (InterruptedException e){
                     Globals.exit("MapHex.loadImages(): "+e);
                  }
               }
            }
            f.dispose();

            pattern= new Image[1];
            pattern[0]= GraphicsUtil.makeSolidHex(Color.red.getRGB(),rowWidths,outlineColor);
         }
         else {
            for (int i=0; i<TerrainTypes.numTypes; i++)
               terrainImages[i]= null;
            pattern= null;
            black= null;
         }
         rectHeight   = 2*rowWidths.length;
         rectWidth    = rowWidths[rowWidths.length-1];
         columnSpacing= rectWidth+rowWidths[0];
         oddRowOffsetX= rectWidth-(rectWidth-rowWidths[0])/2;
         oddRowOffsetY= rowWidths.length;
      }
      initialized= true;
   }

   // static hex grid coordinate and image utility methods
      private static int adjacentRowsOffset[]=       {-2,-1,+1,+2,+1,-1};
      private static int evenRowAdjacentColsOffset[]={ 0, 0, 0, 0,-1,-1};
      private static int oddRowAdjacentColsOffset[]= { 0,+1,+1, 0, 0, 0};
      public static int[] getNeighborCoordinates(int neighbor, int row, int column){
         int rows[]=adjacentRowsOffset;
         int cols[];

         if (row%2==0) cols= evenRowAdjacentColsOffset;
         else cols= oddRowAdjacentColsOffset;

         int tmp[]= new int[2];
         tmp[0]= row+rows[neighbor];
         tmp[1]= column+cols[neighbor];
         return tmp;
      }
      public static int numNeighbors(){return 6;}
/*
      public static int[] newNeighborList(int oldRow, int oldCol, int newRow, int newCol){
         int tmp= new int[6];
         int dr= newRow-oldRow; int dc= newCol-oldCol;
         if (oldRow%2==0){
         }
         else {
         }
         return tmp;
      }
*/

      public static int[] getRowWidths(){return (int[])rowWidths.clone();}
      public static int getWidth(){return rectWidth;}
      public static int getHeight(){return rectHeight;}
      public static int getColumnSpacing(){return columnSpacing;}
      public static int getRowSpacing(){return rectHeight/2;}
      public static Dimension convertPixelSizeToRowColSize(Dimension pixelSize){
         int big= rowWidths[rowWidths.length-1];
         int little= rowWidths[0];
         int rows= (int)(pixelSize.height/rowWidths.length)-1;
         if (rows%2==1) rows--;
         int cols= (int)((pixelSize.width-(big-little)/2)/(big+little));
         return new Dimension(cols,rows);
      }
      public static Dimension convertRowColSizeToPixelSize(Dimension rowColSize){
         int big= rowWidths[rowWidths.length-1];
         int little= rowWidths[0];
         int pixelw= rowColSize.width*(big+little) + (big-little)/2;
         int pixelh= (rowColSize.height+1)*rectHeight/2;      
         return new Dimension(pixelw,pixelh);
      }
      public static Image getTerrainImage(int terrain){return terrainImages[terrain];} 
      public static Image getLandImage(){return terrainImages[TerrainTypes.LAND];} 

      final static int getScreenX(int row, int column){
         if (row%2==0){
            return column*columnSpacing;
         }
         return oddRowOffsetX+column*columnSpacing;
      }

      final static int getScreenY(int row, int column){
         if (row%2==0){
            return row*rectHeight/2;
         }
         return oddRowOffsetY+(row-1)*rectHeight/2;
      }

   public final void drawPattern(Graphics g, int row, int column, int patternIndex, Component c){
      g.drawImage(pattern[patternIndex],getScreenX(row,column),getScreenY(row,column),c);
   }

   public final void drawBlack(Graphics g, int row, int column, Component c){
      g.drawImage(black,getScreenX(row,column),getScreenY(row,column),c);
   }

   public final static void drawPiece(Graphics g, Piece p, int row, int column, Component c){
      g.drawImage(p.getImage(),getScreenX(row,column),getScreenY(row,column),c);
   }

   public final void drawTerrain(Graphics g, int row, int column, Component c){
      g.drawImage(terrainImages[terrainType], getScreenX(row,column),getScreenY(row,column),c);
   }

   public final void draw(Graphics g, int row, int column, Component c){
      if (row%2==0) drawAsEvenRow(g,row/2,column,c);
      else drawAsOddRow(g,(row-1)/2,column,c);
   }

   /** Draws MapHex at pixel coordinates x,y.  This routine must be declared as final so that
     * the compiler will inline it.  It must be <I>fast</I>!
     */
   private final void drawFast(Graphics g, int x, int y, Component c){
      g.drawImage(terrainImages[terrainType],x,y,c);

      if (dh!=null)
         // draw piece or city occupying square.
         g.drawImage(dh.getImage(),x,y,c);
   }

   /** Draws MapHex at row and column indexes, assuming row is even;  This is
       declared as final so that the compiler will inline it;  This routine must 
       be <I>fast</I>! */
   public final void drawAsEvenRow(
      Graphics g, 
      int row_2, // =row/2
      int column, 
      Component c
   ){
      drawFast(g,column*columnSpacing,row_2*rectHeight,c);
   }

   /** Draws MapHex at row and column indexes, assuming row is odd;  This is
       declared as final so that the compiler will inline it;  This routine must 
       be <I>fast</I>! */
   public final void drawAsOddRow(
      Graphics g, 
      int row_1_2, // =(row-1)/2 
      int column, 
      Component c
   ){
      drawFast(g,oddRowOffsetX+column*columnSpacing,oddRowOffsetY+row_1_2*rectHeight,c);
   }

   public boolean detachPiece(Piece p){
      if (dh==null) return false;

      boolean tmp= dh.detachPiece(p);
      if (dh.empty()){
         dh= null;
         clearVisibilityFlags();
      }
      return tmp;
   }

   /** Attach p to this hex. Returns false if square is already occupied */
   public boolean attachPiece(Piece p){
      if (dh!=null){
         if (dh.attachPiece(p)){
            markAsExploredBy(p.getOwner());
            return true;
         }
         return false;
      }

      if (!p.canMoveOver(getTerrain())) return false;

       // Create detailed hex to hold piece.
       if (dh==null){
         dh= new DetailedHex(p);
         markAsExploredBy(p.getOwner());
         return true;
      }

      Globals.exit("MapHex.attachPiece(): this line should never be reached!");
      return false;
   }

   public int getOwner(){
      if (dh==null) return -1;
      else return dh.getOwner();
   }

   // methods relating to exploration of the map and selective visibility (the property that
   // a given player can only see what's going on in squared that his pieces are adjacent to).

      public final void markAsExploredBy(int player){
         // Obviously this scheme allows a maximum of 16 players, given that we
         // need explored and visible status for each player and an int is four bytes long.
         attributes|= (1<<player);
      }
      public final boolean isExploredBy(int player){
         if (player==0) return true; 
         return (attributes&(1<<player))!=0;
      }
      public final void setVisibilityTo(int player, boolean visible){
         // Obviously this scheme allows a maximum of 16 players, given that we
         // need explored and visible status for each player and an int is four bytes long.
         if (visible){
            attributes|= (1<<(player+16));

            // When a fixed piece (like a city) becomes visible to a player, the lastKnownOwner 
            // of the piece to the player may change and hence, must be marked.
            Piece p;
            if ((p=getPiece())!=null){
               if (p.isFixed()){
                  p.markLastKnownOwner(player);
               }
            }
         }
         else 
            attributes&= 0xffffffff^(1<<(player+16));
      }
      /** makes hex invisible to all players. */
      public void clearVisibilityFlags(){attributes&=0x0000ffff;}
      public final boolean isVisibleTo(int player){
         if (player==0) return true; 
         return (attributes&(1<<(player+16)))!=0;
      }
      public boolean[] getVisibleStatus(boolean players[]){
         boolean tmp[]= new boolean[players.length];
         for (int i=0; i<players.length; i++){
            if (players[i]){
               if (isVisibleTo(i)) tmp[i]= true;
            }
            else tmp[i]= false;
         }
         return tmp;
      }
      public void updateLastKnownOwners(){
         Piece p= getPiece();
         if (p==null){
            Globals.exit(
               "MapHex.updateLastKnownOwners(): this shouldn't happen! This method was "+
               "called on an empty hex"
            );
         }
         for (int i=0; i<16; i++){
            if (isVisibleTo(i)){
               p.markLastKnownOwner(i);
            }
         }
      }
}

