//FILE: Piece.java

/* GEMpire (Generic Empire for the Masses), a java package to support internet oriented 
 * empire-like strategy games and associated utilities.
 * Copyright (C) 1998 Christopher L. Marshall
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * My current email is christopherlmarshall@yahoo.com.  My home address is:
 * 10006 Leamoore Lane,
 * Vienna VA, 22181
 * 
 * It is my hope that this modest effort might assist others in producing Java
 * programs to implement a varied host of empire-like games.
 */

package GEMpire;

import java.awt.*;
import java.awt.Toolkit;
import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

public class Piece{
   // Attack types.
   public static int NORMAL_ATTACK = 0;
   public static int BOMBING_ATTACK= 1;

   // Attack result types.
   public static int NORMAL_VICTORY = 0;
   public static int BOMBING_VICTORY= 1;
   public static int DEFEAT         = 2;

   private static PieceDatabase pdb;
   private static int nextPieceId= 0;

   public static void loadPieceDatabase(boolean graphicsEnabled){
      pdb= new PieceDatabase(graphicsEnabled);
   }

   // instance variables
   private int id;
   private int type;
   private int owner; // number of the player that owns the piece.
   private int row, column;
   private int movesLeft;
   private int hitPointsLeft;
   private int range;
   private int producingPiece;
   private int turnsTillPieceProduced;
   private int buildingPiece;
   private int turnsTillPieceBuilt;
   private Vector pieces= null; // used by cities, transports and aircraft carriers to store pieces
   private int lastKnownOwner[]= null;

   private boolean isCarried= false;
   private int currentLoad= 0;
   public String getStatus(){
      return 
         pdb.name(type)+" "+id+" at ("+row+","+column+") "+
         "("+hitPointsLeft+"/"+pdb.hitPoints(type)+") "+
         "Moves "+movesLeft+"/"+pdb.movesPerTurn(type)+" "+
         "Range "+range+"/"+pdb.range(type);         
   }
   public Piece(
      int id,
      int type, 
      int owner, 
      int row, 
      int column,
      int movesLeft, 
      int hitPointsLeft, 
      int range,
      int producingPiece,
      int turnsTillPieceProduced,
      int buildingPiece,
      int turnsTillPieceBuilt
   ){
      this.id= id;
      this.type= type;
      this.owner= owner;
      this.row= row;
      this.column= column;
      this.movesLeft= movesLeft;
      this.hitPointsLeft= hitPointsLeft;
      this.range= range;
      this.producingPiece= producingPiece;
      this.turnsTillPieceProduced= turnsTillPieceProduced;
      this.buildingPiece= buildingPiece;
      this.turnsTillPieceBuilt= turnsTillPieceBuilt;
      if (isFixed()){
         lastKnownOwner= new int[pdb.numPossiblePlayers];
         for (int i=0; i<lastKnownOwner.length;i++){
            lastKnownOwner[i]= -1;
         }
      }
   }
   public Piece(int type, int owner, int row, int column){
      this(
         nextPieceId++,type,owner,row,column,pdb.movesPerTurn(type),pdb.hitPoints(type),
         pdb.range(type),-1,-1,-1,-1
      );
   }
   public Piece(
      int type, int owner, int row, int column, int movesLeft, int hitPointsLeft, 
      int range, int producingPiece, int turnsTillPieceProduced, int buildingPiece, int turnsTillPieceBuilt
   ){
      this(
         nextPieceId++,type,owner,row,column,movesLeft,hitPointsLeft,range,
         producingPiece,turnsTillPieceProduced,buildingPiece,turnsTillPieceBuilt
      );
   }
   public Piece(BufferedReader in) throws IOException, NullPointerException{
      this(in.readLine());
   }

   public Object clone(){
      Piece tmp= new Piece(
         id,type,owner,row,column,movesLeft,hitPointsLeft,range,
         producingPiece,turnsTillPieceProduced,buildingPiece,turnsTillPieceBuilt
      );
      if (lastKnownOwner!=null){
         tmp.lastKnownOwner= (int[])lastKnownOwner.clone();
      }
      return tmp;
   }

   // this toString() is intended for debug purposes.
   public String toString(){
      return "[id: "+id+" type: "+type+" owner:"+owner+
                  " movesLeft:"+movesLeft+
                  " hitPointsLeft:"+hitPointsLeft+
                  " range: "+range+
                  " isCarried: "+isCarried+
                  " currentLoad: "+currentLoad+
                  " row:"+row+" column:"+column+"]";
   }

   /** causes this piece to attack Piece p.  Returns true if attacker wins and false
     * if defender wins.
     */
   public boolean attack1(int terrainA, Piece p, int terrainB){
      if (Globals.attackerAlwaysWins()){
         if (p.isFixed()){
            hitPointsLeft= 0;
            p.changeOwners(getOwner());
         }
         else {
            p.hitPointsLeft= 0;
         }
         return true;
      }
      if (Globals.defenderAlwaysWins()){
         hitPointsLeft= 0;
         return false;
      }

      if (Math.random()<0.5){
         if (p.isFixed()){
            hitPointsLeft= 0;
            p.changeOwners(getOwner());
         }
         else {
            p.hitPointsLeft= 0;
         }
         return true;
      }
      else {
         hitPointsLeft= 0;
         return false;
      }
   }

   private void absorbBombingLoss(){
      double x= Math.random();
System.out.println("Piece.absorbBombingLoss(): "+x);
      if (x<0.5){
         if (producingPiece!=-1){
            turnsTillPieceProduced= pdb.productionTime(type,producingPiece);
System.out.println(toString()+" lost production");
         }
      }
      else {
         if (pieces!=null){
            int n= pieces.size();
            int index;
            do {index= (int)(Math.random()*n);} while (index==n);
System.out.println("losing piece "+index);
            Piece p= (Piece)(pieces.elementAt(index));
            if (!detachPiece(p)){
               System.err.println(
                  "Piece.absorbBombingLoss(): could not remove piece "+p+"\n\rThis shouldn't happen!"
               );
            }
System.out.println(toString()+" lost random contained piece "+p);
         }
      }
      
   }

   private int bombingAttack(Piece p){
      double x= Math.random();
System.out.println("Piece.bombingAttack(): "+x);
System.out.println("Piece.bombingAttack(): successRate="+pdb.bombingSuccessRate(type,p.type));
      if (x<pdb.bombingSuccessRate(type,p.type)){
         p.absorbBombingLoss();
         return BOMBING_VICTORY;
      }
      return DEFEAT;
   }

   public int attack(int terrainA, Piece p, int terrainB){
      if (attackType(p.type)==Piece.BOMBING_ATTACK)
         return bombingAttack(p);

      if (Globals.attackerAlwaysWins()){
         p.hitPointsLeft= 0;
         return NORMAL_VICTORY;
      }
      if (Globals.defenderAlwaysWins()){
         hitPointsLeft= 0;
         return DEFEAT;
      }

      double a= pdb.attackFactor(type,terrainA,p.type,terrainB);
      double b= pdb.defenseFactor(p.type,terrainB,type,terrainA);
      if (a==0||b==0){
         Globals.exit(
            "Piece.attack(): attack or defense factor was zero, this shouldn't happen!\n\r"+
            "p1= "+this+"\n\r"+
            "terrainA= "+terrainA+"\n\r"+
            "p1= "+p+"\n\r"+
            "terrainB= "+terrainB+"\n\r"+
            "attackFactor= "+a+"n\r"+
            "defenseFactor= "+b+"n\r"
         );
      }
      double attackProb= a/(a+b);
      int sdA= pdb.shotDamage(type);
      int sdB= pdb.shotDamage(p.type);
System.out.println("attackFactor: "+a);
System.out.println("defenseFactor: "+b);
System.out.println("shotDamage for attakcer: "+sdA);
System.out.println("shotDamage for defender: "+sdB);

      if (hitPointsLeft<=0||p.hitPointsLeft<=0){
         System.err.println(
           "Piece.attack(): one piece already dead(hp="+
            hitPointsLeft+
            "/"+
            p.hitPointsLeft+
            ").  "+
            "This shouldn't happen!"
         );
      }

      while (hitPointsLeft>0&&p.hitPointsLeft>0){
         double x= Math.random();
System.out.println("x= "+x);
         if (x<attackProb)
            p.hitPointsLeft-= sdA;
         else 
            hitPointsLeft-= sdB;
      }

System.out.println("attacker: "+this);
System.out.println("defender: "+p);

      if (hitPointsLeft>0) return NORMAL_VICTORY;

      return DEFEAT;
   }
   // Methods to pass through Piece Database info.
   public static int numTypes(){return pdb.numPieceTypes();}
   public static String nameOfType(int type){return pdb.name(type);}
   public static Color playerColor(int playerNumber){return pdb.playerColor(playerNumber);}
   public static int numPossiblePlayers(){return pdb.numPossiblePlayers();}
   public boolean canMoveOver(int terrain){return pdb.canMoveOver(type,terrain);}
   /** Does piece have enough movement left for this type of terrain? */
   public boolean canMoveInto(int terrain){
      int tmp= pdb.movementCost(type, terrain);
      return (tmp!=-1)&&(movesLeft>=tmp);
   }
   public boolean canSee(Piece p){
      if (p==null) return false; 
      if (p.owner==owner) return true;
      return pdb.canSee(type,p.type);
   }
   public int attackType(int typeb){return pdb.attackType(type,typeb);}
   public boolean canAttack(int terrainA, Piece p, int terrainB){
      if (movesLeft<=0) return false;
      return pdb.canAttack(type, terrainA, p.getType(), terrainB);
   }
   public static final Image getImage(int type, int owner){
      return pdb.getImage(type,owner);
   }

   public final Image getImage(){return pdb.getImage(type,owner);}
   public boolean isFixed(){return pdb.isFixed(type);}
   public boolean isNeutral(){return pdb.isNeutral(owner);}
   public boolean isAlive(){return hitPointsLeft>0;}
   public int getMovesLeft(){return movesLeft;}
   public void setMovesLeft(int x){movesLeft= x;}
   public int getHitPointsLeft(){return hitPointsLeft;}
   public int getType(){return type;}
   public int getOwner(){return owner;}
   public void resetRange(){
      if (range>=0){
         range= pdb.range(type);
      }
   }
   /** This method is called at the end of a move to an adjacent cell to update range 
     * and movesLeft and to clear any build orders.
     */
   public void checkMove(int terrain, boolean combatOnly, boolean boardingMove){
      int mCost= 1; // the cost of an attack that did not result in a move.
      if (!combatOnly)
         if (boardingMove)
            // here we are setting the cost of boarding moves to 1.
            // I normally make any piece property like this part of the configuration file
            // but I can't imagine how it would be fun to change this so I am just hard coding 
            // it.  So shoot me.
            mCost= 1;
         else
            mCost= pdb.movementCost(type,terrain);

      movesLeft-= mCost;
      if (movesLeft<0){
         System.out.println(
            "Piece.checkMove(): movesLeft is less than zero.  "+
            "This shouldn't happen!"
         );
      }

      // If this piece has a limited range and has exhausted it without 
      // landing in or on some sort of carrier it, kill it.
      if (range>0){
         range-= mCost;
         if (range<0){
            System.out.println(
               "Piece.checkMove(): range is less than zero.  "+
               "This shouldn't happen!"
            );
         }
         if (range==0&&!isCarried){
            hitPointsLeft= 0;
         }
      }
      // Reset any build orders.  We don't want pieces that are supposed to be building something
      // to be able to move as well.
      buildingPiece= -1;
      turnsTillPieceBuilt= -1;
   }
   /** This is called by the PieceMover to make sure the piece can move.  For now this only
     * means that the piece has moves left this turn.  In the future, other considerations may
     * stop the piece from moving.
     */
   public boolean canMove(){return (movesLeft>0);}

   public void markLastKnownOwner(int player){lastKnownOwner[player]= owner;}
   // this is only used for cities (i.e. fixed pieces) that have been defeated in an attack
   public void changeOwners(int newOwner){
      owner= newOwner;
      pieces= null;
      hitPointsLeft= pdb.hitPoints(type);
   }
   public int getId(){return id;}
   public boolean equals(Piece p){return (this.id==p.id);}
   public synchronized void setRowAndColumn(int r, int c){row= r; column= c;}
   public synchronized int[] getRowAndColumn(){
      int tmp[]= new int[2];
      tmp[0]= row; tmp[1]= column;
      return tmp;
   }
   public int getRow(){return row;}
   public int getColumn(){return column;}

   public Piece(String s) throws IOException{
      DelimitedLineReader dlr= new DelimitedLineReader(',');
      String pieceStrings[]= new String[0];

      if (s.charAt(0)=='m'){ // an initial 'm' marks this line as a multiple piece line.
         DelimitedLineReader dlr2= new DelimitedLineReader('|');
         dlr2.setLine(s.substring(1));
         String mainPiece= dlr2.getNextField();
         pieceStrings= dlr2.getFieldsAsArray(); // holds descriptions of pieces carried by this piece.

         // set line with this piece's fields for rest of method.
         dlr.setLine(mainPiece);
      }
      else if (s.charAt(0)=='s'){ // an initial 's' marks this line as a single piece line.
         dlr.setLine(s.substring(1));
      }
      else throw new IOException("badly formed piece string (string should begin with 's' or 'm'): "+s);

      id= dlr.getNextInt();
      type= dlr.getNextInt();
      owner= dlr.getNextInt();
      row= dlr.getNextInt();
      column= dlr.getNextInt();
      movesLeft= dlr.getNextInt();
      hitPointsLeft= dlr.getNextInt();
      range= dlr.getNextInt();
      producingPiece= dlr.getNextInt();
      turnsTillPieceProduced= dlr.getNextInt();
      buildingPiece= dlr.getNextInt();
      turnsTillPieceBuilt= dlr.getNextInt();
      if (isFixed()){
         lastKnownOwner= new int[pdb.numPossiblePlayers];
         for (int i=0; i<lastKnownOwner.length; i++){
            lastKnownOwner[i]= -1;
         }
         Vector tmp= new Vector();
         String stmp= dlr.getNextField();
         int i= 0;
         while (stmp!=null){
            lastKnownOwner[i]= -1;
            try { 
               lastKnownOwner[i]= Integer.parseInt(stmp);
            }
            catch (NumberFormatException e){
            }
            stmp= dlr.getNextField();
            i++;
         }
      }         

      // attach any pieces found above.
      for (int i=0; i<pieceStrings.length; i++){
         if (!attachPiece(new Piece(pieceStrings[i]))){
            System.err.println("Piece(String): could not attach piece in string.  This shouldn't happen!");
         }
      }

      if (id>=nextPieceId) nextPieceId= id+1;
   }

   public void writeCellUpdateTo(PrintWriter out, boolean isVisible, int player, boolean assumeVisible){
      write(out,player,false);
   }
   // manipulation of carried pieces
   public boolean canCarry(Piece p){return pdb.canCarry(type,p.type);}
   public boolean containsPieces(){if (pieces==null) return false; return true;}

   public boolean attachPiece(Piece p){
      if (pdb.canCarry(type,p.type)&&owner==p.owner){
         int newLoad= currentLoad+pdb.carryingCost(type,p.type);
         if (newLoad<=pdb.carryingCapacity(type)){
            currentLoad= newLoad;
            if (pieces==null) pieces= new Vector();
            pieces.addElement(p);
            p.row= row;
            p.column= column;
            p.isCarried= true;

            // move any carried pieces from p to 'this'.
            Piece pTmp;
            while ((pTmp=p.getPiece())!=null){
               p.detachPiece(pTmp);
               if (!attachPiece(pTmp)){
                  Globals.exit("Piece.attachPiece(): this shouldn't happen!  Failed to attach "+pTmp);
               }
            }
            return true;
         }
      }
      return false;
   }
   public boolean detachPiece(Piece p){
      Piece tmp= getPiece(p.id);
      if (tmp==null) return false;
      pieces.removeElement(tmp);
      if (pieces.size()==0) pieces= null;
      p.isCarried= false;
      currentLoad-= pdb.carryingCost(type,p.type);
      if (currentLoad<0){
         System.err.println("Piece.detachPiece(): load went below zero.  This shouldn't happen!");
      }
      return true;
   }
   public Piece getPiece(int id){
      if (pieces==null) return null;
      Enumeration e= pieces.elements();
      while (e.hasMoreElements()){
         Piece tmp= (Piece)(e.nextElement());
         if (tmp.id==id){
            tmp.setRowAndColumn(row,column);
            return tmp;
         }
      }
      return null;
   }
   public Piece getPiece(){
      if (pieces==null) return null;
      if (pieces.size()>0){
         Piece tmp= (Piece)(pieces.elementAt(0));
         pieces.removeElement(tmp);
         pieces.addElement(tmp);
         tmp.setRowAndColumn(row,column);
         return tmp;
      }
      return null;
   }
   public Vector getPieces() {return pieces;}

   /** This method is only called by write() and handles the special case of a piece that does not
     * carry other pieces.
     */
   private void simpleWrite(PrintWriter out, int player, boolean contained){
      out.print("s");
      if (player==0){ // Write all information for the null player
         out.print(
            id+","+type+","+owner+","+row+","+column+","+movesLeft+","+hitPointsLeft+","+range+","+
            producingPiece+","+turnsTillPieceProduced+","+buildingPiece+","+turnsTillPieceBuilt
         );
         if (isFixed()){
            for (int i=0; i<lastKnownOwner.length; i++){
               out.print(","+lastKnownOwner[i]);
            }
         }
      }
      else if (owner==player){
         out.print(
            id+","+type+","+owner+","+row+","+column+","+movesLeft+","+hitPointsLeft+","+range+","+
            producingPiece+","+turnsTillPieceProduced+","+buildingPiece+","+turnsTillPieceBuilt
         );
      }
      else if (isFixed()){
         out.print(id+","+type+","+lastKnownOwner[player]+","+row+","+column+",-1,-1,-1,-1,-1,-1,-1");
      }
      else {
         out.print(id+","+type+","+owner+","+row+","+column+",-1,-1,-1,-1,-1,-1,-1");
      }

      if (!contained) out.println();
   }

   /** This method is used to write all of the information about a Piece to a file.  It has been
     * carefully constructed to only write the information that a given player is allowed to know
     * about a piece.  Even if the player owns the piece, she is not allowed to know certain information
     * that is maintained by pieces, like who owned the piece the last time each player spotted it
     * (the field lastKnownOwner[] for fixed pieces, for example).  When the player argument is -1,
     * the null player, all of the information about the piece is written.
     */
   public void write(PrintWriter out, int player, boolean contained){
      if (pieces!=null){ // this piece is carrying other pieces
         if (player==0){ // print everything
            out.print("m");
            out.print(
               id+","+type+","+owner+","+row+","+column+","+movesLeft+","+hitPointsLeft+","+range+","+
               producingPiece+","+turnsTillPieceProduced+","+buildingPiece+","+turnsTillPieceBuilt
            );
            if (isFixed()){
               for (int i=0; i<lastKnownOwner.length; i++){
                  out.print(","+lastKnownOwner[i]);
               }
            }
            out.print("|");

            int n= pieces.size();
            for (int i=0; i<n-1; i++){
               ((Piece)(pieces.elementAt(i))).write(out,player,true);
               out.print("|");
            }
            ((Piece)(pieces.elementAt(n-1))).write(out,player,true);
            out.println();
         }
         else if (owner==player){ //print everything but info about other players.
            out.print("m");
            out.print(
               id+","+type+","+owner+","+row+","+column+","+movesLeft+","+hitPointsLeft+","+range+","+
               producingPiece+","+turnsTillPieceProduced+","+buildingPiece+","+turnsTillPieceBuilt
            );
            out.print("|");

            int n= pieces.size();
            for (int i=0; i<n-1; i++){
               ((Piece)(pieces.elementAt(i))).write(out,player,true);
               out.print("|");
            }
            ((Piece)(pieces.elementAt(n-1))).write(out,player,true);
            out.println();
         }
         else { // player is not owner.  write piece as if it does not carry any other pieces.
           simpleWrite(out,player,contained);
         }
      }
      else {
         simpleWrite(out,player,contained);
      }
   }
   public void write(PrintWriter out, int player){write(out,player,false);}
   public void write(PrintWriter out){write(out,0,false);}

   public void endTurn(){
      // implement fuel drain.
      if (range>=0){ // if piece has limited range
         if (movesLeft>0){
            int tmp= range-movesLeft;
            if (tmp<=0){
               range= 0;
               hitPointsLeft= 0;
            }
            else range= tmp;
         }
      }
   }
   public Piece refresh(){
      boolean movedThisTurn= (movesLeft<pdb.movesPerTurn(type));

      // refresh movement ability, repair damage
      movesLeft= pdb.movesPerTurn(type);
      if (isCarried){
         resetRange();
         if (!movedThisTurn)
            if (hitPointsLeft<pdb.hitPoints(type)){
               hitPointsLeft++;
            }
      }

      // Check for new piece production.
      if (turnsTillPieceProduced>=0){
         turnsTillPieceProduced--;
         if (turnsTillPieceProduced==0){
            if (producingPiece==-1)
               Globals.exit(this+" producing piece -1!  This shouldn't happen!");
            turnsTillPieceProduced= pdb.productionTime(type,producingPiece);

            Piece tmp= new Piece(producingPiece,owner,row,column);
            if (!attachPiece(tmp)){
               Globals.exit(
                  "Piece.refresh("+
                   this+
                   "): Couldn't attach piece just produced: "+
                   tmp+
                   "\n\rThis shouldn't happen!"
               );
            }
         }
      }

      // Refresh any carried pieces.
      if (pieces!=null){
         int N= pieces.size();
         for (int i=0; i<N; i++){
            ((Piece)(pieces.elementAt(i))).refresh();
         }
      }

      // Check for piece being built.  The new piece replaces the old one (like an 
      // infantry building an airbase)
      if (turnsTillPieceBuilt>=0){
         turnsTillPieceBuilt--;
         if (turnsTillPieceBuilt==0){
            if (buildingPiece==-1)
               Globals.exit(this+" building piece -1!  This shouldn't happen!");
            turnsTillPieceBuilt= -1;

            return new Piece(buildingPiece,owner,row,column);
         }
      }

if (isFixed()){
   System.out.println("Piece.refresh()");
   System.out.println("turnsTillPieceProduced= "+turnsTillPieceProduced);
   System.out.println("producingPiece= "+producingPiece);
}
      return null;
   }

   public int getProduction(){return producingPiece;}
   public boolean setProduction(int ptype){
      if (ptype==producingPiece) return true;
      if (ptype==-1){
         producingPiece= -1;
         return true;
      }
      if (pdb.canProduce(type,ptype)){
         producingPiece= ptype;
         turnsTillPieceProduced= pdb.productionTime(type,ptype);
         return true;
      }
      return false;
   }
   public int getBuildType(){return buildingPiece;}
   public boolean setBuildType(int ptype){
      if (ptype==buildingPiece){
         return true;
      }
      if (ptype==-1){
         producingPiece= -1;
         return true;
      }
      if (pdb.canBuild(type,ptype)){
         buildingPiece= ptype;
         turnsTillPieceBuilt= pdb.buildTime(type,ptype);
         return true;
      }
      return false;
   }

   public int getBuildType(Frame f){
      // Count the number of different types this piece can build.
      int count= 0;
      int n= pdb.numPieceTypes();
      for (int i=0; i<n; i++)
         if (pdb.buildTime(type,i)!=-1)
            count++;
      if (count==0) return -1;

System.out.println("Point 1"); System.out.flush();
      // Form lists of the names, types, and build times of all of the types this piece
      // can build.  Also produce a report (in a StringWriter) of these that will be 
      // displayed to the user in the selector dialog.
      String nameList[]= new String[count+1];
      int typeList[]= new int[count+1];
      int timeToBuildList[]= new int[count+1];
      int j=1;
      StringWriter sw= new StringWriter();
      PrintWriter pw= new PrintWriter(sw);
      pw.println("Name\tBuild Time\tTimeTillPieceBuilt");
      nameList[0]= "Nothing";
      typeList[0]= -1;
      timeToBuildList[0]= -1;
      pw.println(nameList[j]+"\t"+timeToBuildList[j]+"\t*");
      int curValue= 0;
System.out.println("Point 2"); System.out.flush();
      for (int i=0; i<n; i++){
         if (pdb.buildTime(type,i)!=-1){
            if (buildingPiece==i)
               curValue= j;
            typeList[j]= i;
            nameList[j]= pdb.name(i);
            timeToBuildList[j]= pdb.buildTime(type,i);
            if (buildingPiece==i)
               pw.println(nameList[j]+"\t"+timeToBuildList[j]+"\t"+turnsTillPieceBuilt);
            else
               pw.println(nameList[j]+"\t"+timeToBuildList[j]+"\t*");
            j++;
         }
      }
      pw.flush();

System.out.println("Point 3"); System.out.flush();
System.out.println("current value= "+curValue); System.out.flush();
      // Get an index from the user into the list of types that can be built by this piece.
      SimpleSelectorDialog buildDialog= new SimpleSelectorDialog(
         f,
         "PieceBuildDialog",
         nameList,
         sw.toString(),
         15,
         40,
         curValue
      );
      pw.close();
      try {sw.close();} catch (IOException e){}
      buildDialog.dispose();
      int selectedIndex= buildDialog.getSelectedIndex();

      // Set the build type locally and return the type selected.  I really don't like this
      // approach, but I'm doing it anyway.  What should happen is that the server sets the
      // build type remotely and sends an update to the client, overwriting the old piece.
      // This is how the local piece should have its build type changed, or any other 
      // parameter changed for that matter.
      setBuildType(typeList[selectedIndex]);
      return typeList[selectedIndex];
   }

   public void displayProductionDialog(Frame f){
      boolean canNotProduceAnything= true;
      int initialProductionIndex= 0;

      int numPieceTypes= numTypes();
      int produciblePieceTypes[]= new int[numPieceTypes];
      Vector tmp= new Vector();
      tmp.addElement("Nothing");
      StringWriter sw= new StringWriter();
      PrintWriter pw= new PrintWriter(sw);
      pw.println("Type\tTurns To Produce\tTurns Left");
      int j=0;
      for (int i=0; i<numPieceTypes; i++){
         if (pdb.canProduce(type,i)){
            canNotProduceAnything= false;
            produciblePieceTypes[j++]= i;
            tmp.addElement(pdb.name(i));
            if (producingPiece==i){
               pw.println(pdb.name(i)+"\t"+pdb.productionTime(type,i)+"\t"+turnsTillPieceProduced);
               initialProductionIndex= j;
            }
            else 
               pw.println(pdb.name(i)+"\t"+pdb.productionTime(type,i)+"\t*");
         }
      }
      pw.flush();

      if (canNotProduceAnything) return;

      String pnames[]= new String[tmp.size()];
      for (int i=0; i<pnames.length; i++) pnames[i]= (String)(tmp.elementAt(i));
      SimpleSelectorDialog ppd= new SimpleSelectorDialog(
         f,
         "PieceProductionDialog",
         pnames,
         sw.toString(),
         15,
         40,
         initialProductionIndex
      );
      pw.close();
      try {sw.close();} catch (IOException e){}
      ppd.dispose();

      int selectedIndex= ppd.getSelectedIndex()-1;
      if (selectedIndex==-1)
         setProduction(-1);
      else
         setProduction(produciblePieceTypes[selectedIndex]);

   }
}

