//FILE: PieceInfo.java

/* GEMpire (Generic Empire for the Masses), a java package to support internet oriented 
 * empire-like strategy games and associated utilities.
 * Copyright (C) 1998 Christopher L. Marshall
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * My current email is christopherlmarshall@yahoo.com.  My home address is:
 * 10006 Leamoore Lane,
 * Vienna VA, 22181
 * 
 * It is my hope that this modest effort might assist others in producing Java
 * programs to implement a varied host of empire-like games.
 */

package GEMpire;

import java.io.*;
import java.util.Vector;

/** This class is used by PieceInfo to hold attack or defense factor modifications that take into 
  * account piece and terrain type (say you want to forbid destroyers from attacking fighters over 
  * land, or you want to double the defense factor of an infantry in a mountain cell against all 
  * attackers).  Any number of these can apply to a given attacking or defending piece type.
  */
class CombatModifier{
   static DelimitedLineReader dlr= new DelimitedLineReader(',');
   static final int ATTACK = 0;
   static final int DEFENSE= 1;

   int type;

   int terrainTypeA;
   int pieceTypeB;
   int terrainTypeB;
   double modifier;

   CombatModifier(String line, int numPieceTypes, int numTerrainTypes) throws IOException{
      dlr.setLine(line);
      type        = dlr.getNextInt();
      terrainTypeA= dlr.getNextInt();
      pieceTypeB  = dlr.getNextInt();
      terrainTypeB= dlr.getNextInt();
      modifier=     dlr.getNextDouble();

      if (type!=ATTACK&&type!=DEFENSE) 
         throw new IOException("invalid CombatModifier Type");

      if (
         (terrainTypeA<-1||terrainTypeA>=numTerrainTypes)||
         (terrainTypeB<-1||terrainTypeB>=numTerrainTypes)
      )
         throw new IOException("invalid terrain type encountered");

      if (pieceTypeB<-1||pieceTypeB>numPieceTypes)
         throw new IOException("invalid terrain type encountered");

   }

   final double value(){return modifier;}
   final boolean appliesTo(int terrainA, int pieceB, int terrainB){
      if (
         (terrainA==terrainTypeA||terrainTypeA==-1)&&
         (pieceB  ==pieceTypeB  ||pieceTypeB  ==-1)&&
         (terrainB==terrainTypeB||terrainTypeB==-1)
      )
         return true;
      return false;
   }
   final int getType(){return type;}
   public String toString(){
      StringWriter sw= new StringWriter();
      PrintWriter  pw= new PrintWriter(sw);

      if (type==ATTACK) pw.println("ATTACK Modifier");
      else pw.println("DEFENSE Modifier");
      pw.println("terrainTypeA= "+terrainTypeA);
      pw.println(  "pieceTypeB= "+pieceTypeB);
      pw.println("terrainTypeB= "+terrainTypeB);
      pw.flush();
      return sw.toString();
   }
}

/**  This type holds all of the properties info for a given piece type.
  */
class PieceInfo{
   String name;
   int movesPerTurn;
   int range;
   int damage;
   int shotDamage;
   boolean isFixed;
   int movementCost[];
   int carryingCapacity;
   int carryingCost[];
   boolean canSee[];
   double attackFactors[];
   double defenseFactors[];
   int productionTime[];
   int buildTime[];

   CombatModifier attackModifiers[];
   CombatModifier defenseModifiers[];

   static boolean isPieceDescription(String line){
      int tmp= line.indexOf("piece");
      return (tmp==0);
   }

   static int[] convertToIntArray(String s[]) throws IOException{
      int ia[]= new int[s.length];
      for (int i=0; i<s.length; i++){
         try {
            if (s[i]==null) 
               ia[i]= -1;
            else
               ia[i]= Integer.parseInt(s[i].trim());
         }
         catch (NumberFormatException e){
            throw new IOException("PieceInfo.convertToIntArray()");
         }
      }
      return ia;
   }

   PieceInfo(String line, int numPieceTypes) throws IOException{

      String tmp,field,slist[];
      int x[],y;

      DelimitedLineReader dlr1= new DelimitedLineReader('\t');
      DelimitedLineReader dlr2= new DelimitedLineReader(',');
      dlr1.setLine(line);

      // check line marker
      if (!dlr1.getNextField().equals("piece")) 
         throw new IOException("PieceInfo(): line is not a valid piece description: "+line);

      // type field is read but not used.
      dlr1.getNextField();

      name= dlr1.getNextField();
      movesPerTurn= dlr1.getNextInt();
      range= dlr1.getNextInt();
      damage= dlr1.getNextInt();
      shotDamage= dlr1.getNextInt();
      isFixed= (dlr1.getNextInt()==1);

      // Read ivar productionTime[].

      // Read Production Time
         // parse properties field into string array.
         slist= IOUtil.parsePropertiesList(dlr1.getNextField().trim(),0,numPieceTypes-1,',','/');
         if (slist==null) throw new IOException("productionTime field malformed in piece info file");
         // Convert string array to int array with 'null' becoming -1.
         productionTime= convertToIntArray(slist);

      // Read Build Time
         // parse properties field into string array.
         slist= IOUtil.parsePropertiesList(dlr1.getNextField().trim(),0,numPieceTypes-1,',','/');
         if (slist==null) throw new IOException("buildTime field malformed in piece info file");
         // Convert string array to int array with 'null' becoming -1.
         buildTime= convertToIntArray(slist);

      // Read Movement Cost
         // parse properties field into string array.
         slist= IOUtil.parsePropertiesList(dlr1.getNextField().trim(),0,TerrainTypes.numTypes-1,',','/');
         if (slist==null) throw new IOException("movementCost field malformed in piece info file");
         // Convert string array to int array with 'null' becoming -1.
         movementCost= convertToIntArray(slist);

      // Read ivar carryingCapacity.
      carryingCapacity= dlr1.getNextInt();

      // Read ivar carryingCost[].
      slist= IOUtil.parsePropertiesList(dlr1.getNextField().trim(),0,numPieceTypes-1,',','/');
      if (slist==null) throw new IOException("productionTime field malformed");
      // Convert string array to int array with 'null' becoming -1.
      carryingCost= convertToIntArray(slist);

      // Read ivar canSee[].
      canSee= new boolean[numPieceTypes];
      for (int i=0; i<numPieceTypes; i++) canSee[i]= false;
      tmp= dlr1.getNextField().trim();
      if (!tmp.equals("")){
         dlr2.setLine(tmp);
         x= dlr2.getNextInt2();
         while (x!=null){
            if (x[0]<0||x[0]>=numPieceTypes){
               throw new IOException(
                  "PieceInfo(): invalid piece type read from PieceDatabase file in canSee column: "+x[0]
               );
            }
            canSee[x[0]]= true;
            x= dlr2.getNextInt2();
         }
      }

      // Read AttackFactors
      attackFactors= new double[numPieceTypes];
      tmp= dlr1.getNextField().trim();
      dlr2.setLine(tmp);
      for (int i=0; i<numPieceTypes; i++){
         attackFactors[i]= dlr2.getNextDouble();
      }

      // Read DefenseFactors
      defenseFactors= new double[numPieceTypes];
      tmp= dlr1.getNextField().trim();
      dlr2.setLine(tmp);
      for (int i=0; i<numPieceTypes; i++){
         defenseFactors[i]= dlr2.getNextDouble();
      }

      // Read Attack/Defense Modifiers
      CombatModifier cm;
      Vector vAttackModifiers = new Vector();
      Vector vDefenseModifiers= new Vector();
      tmp= dlr1.getNextField();
      while (tmp!=null){
         tmp= tmp.trim();
         if (!tmp.equals("")){
            try {cm= new CombatModifier(tmp,numPieceTypes,TerrainTypes.numTypes);}
            catch (IOException e){
               throw new IOException(
                  "PieceInfo(): could not read ComabtModifier from field("+tmp+") in line:"+line+"\n\r"+e
              );
            }

            if (cm.getType()==CombatModifier.ATTACK)
               vAttackModifiers.addElement(cm);
            else
               vDefenseModifiers.addElement(cm);
         }
         tmp= dlr1.getNextField();
      }

      attackModifiers = new CombatModifier[ vAttackModifiers.size()];
      for (int i=0; i<attackModifiers.length; i++){
         attackModifiers[i]= (CombatModifier)(vAttackModifiers.elementAt(i));
      }

      defenseModifiers= new CombatModifier[vDefenseModifiers.size()];
      for (int i=0; i<defenseModifiers.length; i++){
         defenseModifiers[i]= (CombatModifier)(vDefenseModifiers.elementAt(i));
      }
   }

   double defenseFactor(int terrainA, int pieceB, int terrainB){
      double value= defenseFactors[pieceB];

      for (int i=0; i<defenseModifiers.length; i++){
         if (defenseModifiers[i].appliesTo(terrainA,pieceB,terrainB)){
            value*= defenseModifiers[i].value();
         }
      }

      return value;
   }

   double attackFactor(int terrainA, int pieceB, int terrainB){
      double value= attackFactors[pieceB];

      for (int i=0; i<attackModifiers.length; i++){
         if (attackModifiers[i].appliesTo(terrainA,pieceB,terrainB)){
            value*= attackModifiers[i].value();
         }
      }

      return value;
   }

   boolean canAttack(int terrainA, int pieceB, int terrainB){
      if (attackFactors[pieceB]==0) return false;

      for (int i=0; i<attackModifiers.length; i++){
         if (attackModifiers[i].appliesTo(terrainA,pieceB,terrainB)){
            if (attackModifiers[i].value()==0) return false;
         }
      }

      return true;
   }

   public String toString(){
      StringWriter sw= new StringWriter();
      PrintWriter pw= new PrintWriter(sw);
      pw.println("name\t"+name);
      pw.println("MovesPerTurn\t"+movesPerTurn);
      pw.println("Range\t"+range);
      pw.println("Damage\t"+damage);
      pw.println("ShotDamage\t"+shotDamage);

      pw.print("MovementCost"); 
      for (int i=0; i<movementCost.length; i++)  pw.println("\t"+i+"\t"+movementCost[i]);

      pw.println("CarryingCost\t");
      for (int i=0; i<carryingCost.length; i++)  pw.println("\t"+i+"\t"+carryingCost[i]);

      pw.println("CarryingCapacity\t"+carryingCapacity);

      pw.println("CanSee\t");
      for (int i=0; i<canSee.length; i++)  pw.println("\t"+i+"\t"+canSee[i]);

      pw.println("AttackFactors\t");
      for (int i=0; i<attackFactors.length; i++)  pw.println("\t"+i+"\t"+attackFactors[i]);

      pw.println("DefenseFactors\t");
      for (int i=0; i<defenseFactors.length; i++)  pw.println("\t"+i+"\t"+defenseFactors[i]);

      pw.println("ProductionTime\t");
      for (int i=0; i<productionTime.length; i++){
         pw.print(" "+productionTime[i]);
      }
      pw.println();

      pw.println("AtttackModifiers");
      for (int i=0; i<attackModifiers.length; i++){
         pw.print(" "+attackModifiers[i]);
      }
      pw.println();

      pw.println("Defense Modifiers");
      for (int i=0; i<defenseModifiers.length; i++){
         pw.print(" "+defenseModifiers[i]);
      }
      pw.println();

      pw.flush();
      return sw.toString();
   }

   public void reportOnCombatProperties(PrintWriter out, PieceInfo pi[]){
      String tname[]={"sea","land"};
      out.println("PieceA\tTerrainA\tPieceB\tTerrainB\tDefense\tattack\tcan");
      for (int i=0; i<movementCost.length; i++){
         for (int k=0; k<movementCost.length; k++){
            for (int j=0; j<carryingCost.length; j++){
               if (attackFactor(i,j,k)!=0||defenseFactor(i,j,k)!=0)
                  out.println(
                     name+"\t"+
                     TerrainTypes.names[i]+"\t"+
                     pi[j].name+"\t"+
                     TerrainTypes.names[k]+"\t"+
                     defenseFactor(i,j,k)+"\t"+
                     attackFactor(i,j,k)+"\t"+
                     canAttack(i,j,k)
               );
            }
         }
      }
   }


   public static void main(String args[]) throws IOException{
      String inputLine="piece	0	city	0	-1	1	1	1	0/-1,1/6,2/12,3/12,4/12,5/-1,6/30,7/24,8/24,9/48,10/36,11/60		0,1	1	1/0,2/0,3/0,4/0,5/0,6/0,7/0,8/0,9/0,10/0,11/0	0,1,2,3,4,5,6,7,9,10,11	0,0,0,0,0,0,0,0,0,0,0,0	0,2,2,2,2,0,0,0,0,0,0,0";
      PieceInfo pi= new PieceInfo(inputLine,12);
      System.out.print(pi.toString());
   }
}






