//FILE: PieceMover.java

/* GEMpire (Generic Empire for the Masses), a java package to support internet oriented 
 * empire-like strategy games and associated utilities.
 * Copyright (C) 1998 Christopher L. Marshall
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * My current email is christopherlmarshall@yahoo.com.  My home address is:
 * 10006 Leamoore Lane,
 * Vienna VA, 22181
 * 
 * It is my hope that this modest effort might assist others in producing Java
 * programs to implement a varied host of empire-like games.
 */

package GEMpire;

public class PieceMover{
   boolean graphicsEnabled;
   Map m;

   PieceMover(Map m){
      this(m,true);
   }
   PieceMover(Map m, boolean graphicsEnabled){
      this.m= m;
      this.graphicsEnabled= graphicsEnabled;
   }

   private boolean simpleMove(Piece p, int row, int column){
      int player= p.getOwner();
      int targetTerrain= m.getTerrain(row,column);
      boolean boardingMove=false;

      if (graphicsEnabled) m.deactivatePiece();

      // Make sure piece has enough movement points left for this terrain type.
      if (!p.canMoveInto(targetTerrain)) {
         // this section is about one piece boarding another
         Piece p2= m.getPieceAt(row,column);
         if (p2==null) return false;
         if (p2.getOwner()!=p.getOwner() || !p2.canCarry(p)) return false;
         boardingMove= true;
      }

      if (!m.detachPiece(p)){
         System.out.println("PieceMover.simpleMove(): Couldn't detach piece from map;  this shouldn't happen!");
         return false;
      }

      int old_rc[]= p.getRowAndColumn();
      p.setRowAndColumn(row,column);
       if (!m.attachPiece(p)){
         p.setRowAndColumn(old_rc[0],old_rc[1]);
         if (!m.attachPiece(p)){
            Globals.exit(
               "PieceMover.simpleMove(): Couldn't reattach "+p+" to map.  This shouldn't happen!"
            );
         }
         return false;
      }

      m.markSurroundingCellsAsExploredBy(row,column,player);
      m.markVisibilityAround(row,column);
      m.determineVisibilityOfCellsSurrounding(old_rc[0],old_rc[1],player);
      m.markVisibilityOfCellToNeighbors(row,column);

      if (graphicsEnabled){
         m.activatePiece(row,column);
         m.drawCell(old_rc[0],old_rc[1]);
      }

      // Handle case where piece runs out of gas and dies (i.e. planes that don't land).
      p.checkMove(targetTerrain,false,boardingMove);
      if (!p.isAlive())
         if (!m.detachPiece(p)){
            Globals.exit(
               "PieceMover.simpleMove(): couldn't detach piece that exhausted its "+
               "range from map.  This shouldn't happen!"
            );
         }

      return true;
   }

   private void handleBombingVictory(Piece p, Piece p2, int player2, int row, int col){
      p.checkMove(-1,true,false);
   }

   private void handleAggressorVictory(Piece p, Piece p2, int player2, int row, int col){
      int playerNumber= p.getOwner();
      int old_rc[]= p.getRowAndColumn();

      if (p2.isFixed()){
         // Victor piece is removed from map to garrison fixed piece.
         if (!m.detachPiece(p)){
            System.out.println(
               "PieceMover.handleAggressorVictory(): could not remove "+p+" from map to garrison "
               +p2
            );
         }

         m.updateLastKnownOwners(p2);

         // to the victor go the spoils
         p2.changeOwners(playerNumber);

         // update visibility for loser
         m.determineVisibilityOfCellsSurrounding(row,col,player2);

         // update visibility for victor
         m.markSurroundingCellsAsExploredBy(row,col,playerNumber);
         m.markVisibilityAround(row,col);
         m.determineVisibilityOfCellsSurrounding(old_rc[0],old_rc[1],playerNumber);
         m.markVisibilityOfCellToNeighbors(row,col);         
      }
      else if (m.detachPiece(p2)){
         // Move victor into new square
         if (!simpleMove(p,row,col)){
            // The '-1' terrain value is for a combat only move, terrain doesn't matter.
            p.checkMove(-1,true,false); 
            System.out.println(
               "PieceMover.handleAggressorVictory(): can't move piece to conquered cell!"
            );
         }
         // update visibility for loser
            // update visibility around cell formerly occupied by p2
            m.determineVisibilityOfCellsSurrounding(row,col,p2.getOwner());
      }
      else 
         System.out.println(
            "PieceMover.handleAggressorVictory(): Couldn't detach piece "+p2+" from map!  "+
            "This shouldn't happen!"
         );
   }

   private void handleDefenderVictory(Piece p, Piece p2, int row, int col){
      int old_rc[]= p.getRowAndColumn();

      if (!m.detachPiece(p)){
         System.out.println(
            "EmpireServer.movePieceTo(): Couldn't detach piece "+p+" from map!  "+
            "This shouldn't happen!"
         );
      }

      m.determineVisibilityOfCellsSurrounding(old_rc[0],old_rc[1],p.getOwner());
   }

   /** This method is called by the rest of the package to handle piece moves.  Most of the code
     * in it deals with whether combat takes place or not and calls private methods to handle the
     * details.  The private method, simpleMove() handles the most frequent case (a piece move into 
     * a friendly cell or unoccupied cell).
     */

   public MovementReport movePieceTo(Piece p, int row, int col, boolean players[]){
      MovementReport mreport= new MovementReport(m,p,row,col,players);
      int player= p.getOwner();

      if (m.withinBounds(row,col)&&p.canMove()){
         int old_rc[]= p.getRowAndColumn();
         if (graphicsEnabled) m.deactivatePiece();
         Piece p2= m.getPieceAt(row,col);

         if (p2!=null){
            int player2= p2.getOwner();
            if (p2.getOwner()!=p.getOwner()){
               int terrainA= m.getTerrain(old_rc[0],old_rc[1]);
               int terrainB= m.getTerrain(row,col);
               if (p.canAttack(terrainA,p2,terrainB)){
                  int attackResult= p.attack(terrainA,p2,terrainB);
                  mreport.setCombatOccurred();

                  if (attackResult==Piece.NORMAL_VICTORY)
                     handleAggressorVictory(p,p2,player2,row,col);
                  else if (attackResult==Piece.BOMBING_VICTORY)
                     handleBombingVictory(p,p2,player2,row,col);
                  else if (attackResult==Piece.DEFEAT)
                     handleDefenderVictory(p,p2,row,col);
                  else {
                     System.err.println("PieceMover.movePieceTo(): unknown attack result= "+attackResult);
                     mreport.storeResults(m,players);
                     return mreport;
                  }
               }
               else {
                  mreport.setMoveOccurred(false);
                  mreport.storeResults(m,players);
                  return mreport;
               }
            }
            else{
               // this case occurs when one piece boards another
               mreport.setMoveOccurred(simpleMove(p, row, col));
               mreport.storeResults(m,players);
               return mreport;
            }
         }
         else{
            mreport.setMoveOccurred(simpleMove(p, row, col));
            mreport.storeResults(m,players);
            return mreport;
         }

         if (graphicsEnabled){
            m.drawCell(old_rc[0], old_rc[1]);
            m.drawCell(row,col);
         }
         mreport.setMoveOccurred(true);
         mreport.storeResults(m,players);
         return mreport;
      }
      mreport.setMoveOccurred(false);
      mreport.storeResults(m,players);
      return mreport;
   }

   public void movePieceDown(Piece p){
      if (p!=null){
         int row= p.getRow();
         int column= p.getColumn();
         movePieceTo(p,row+2,column,new boolean[0]);
      }
   }

   public void movePieceUp(Piece p){
      if (p!=null){
         int row= p.getRow();
         int column= p.getColumn();
         movePieceTo(p,row-2,column,new boolean[0]);
      }
   }
   public void movePieceUpAndRight(Piece p){
      if (p!=null){
         int row= p.getRow();
         int column= p.getColumn();
         if (row%2==0){
            movePieceTo(p,row-1,column,new boolean[0]);
         }
         else {
            movePieceTo(p,row-1,column+1,new boolean[0]);
         }
      }
   }
   public void movePieceDownAndRight(Piece p){
      if (p!=null){
         int row= p.getRow();
         int column= p.getColumn();
         if (row%2==0){
            movePieceTo(p,row+1,column,new boolean[0]);
         }
         else {
            movePieceTo(p,row+1,column+1,new boolean[0]);
         }
      }
   }
   public void movePieceDownAndLeft(Piece p){
      if (p!=null){
         int row= p.getRow();
         int column= p.getColumn();
         if (row%2==0){
            movePieceTo(p,row+1,column-1,new boolean[0]);
         }
         else {
            movePieceTo(p,row+1,column,new boolean[0]);
         }
      }
   }
   public void movePieceUpAndLeft(Piece p){
      if (p!=null){
         int row= p.getRow();
         int column= p.getColumn();
         if (row%2==0){
            movePieceTo(p,row-1,column-1,new boolean[0]);
         }
         else {
            movePieceTo(p,row-1,column,new boolean[0]);
         }
      }
   }
}
